#!/usr/bin/env python
# Copyright 2014-2021 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Different FCI solvers are implemented to support different type of symmetry.
                    Symmetry
File                Point group   Spin singlet   Real hermitian*    Alpha/beta degeneracy
direct_spin0_symm   Yes           Yes            Yes                Yes
direct_spin1_symm   Yes           No             Yes                Yes
direct_spin0        No            Yes            Yes                Yes
direct_spin1        No            No             Yes                Yes
direct_uhf          No            No             Yes                No
direct_nosym        No            No             No**               Yes

*  Real hermitian Hamiltonian implies (ij|kl) = (ji|kl) = (ij|lk) = (ji|lk)
** Hamiltonian is real but not hermitian, (ij|kl) != (ji|kl) ...
'''

import sys
import ctypes
import numpy
import numpy as np
from pyscf import ao2mo
from pyscf import lib
from pyscf.lib import logger
from pyscf import symm
from pyscf.scf.hf_symm import map_degeneracy
from pyscf.fci import cistring
from pyscf.fci import direct_spin1
from pyscf.fci import addons
from pyscf.fci.spin_op import contract_ss
from pyscf.fci.addons import _unpack_nelec
from pyscf import __config__

libfci = direct_spin1.libfci

TOTIRREPS = 8

# Note eri is NOT the 2e hamiltonian matrix, the 2e hamiltonian is
# h2e = eri_{pq,rs} p^+ q r^+ s
#     = (pq|rs) p^+ r^+ s q - (pq|rs) \delta_{qr} p^+ s
# so eri is defined as
#       eri_{pq,rs} = (pq|rs) - (1/Nelec) \sum_q (pq|qs)
# to restore the symmetry between pq and rs,
#       eri_{pq,rs} = (pq|rs) - (.5/Nelec) [\sum_q (pq|qs) + \sum_p (pq|rp)]
# Please refer to the treatment in direct_spin1.absorb_h1e
def contract_2e(eri, fcivec, norb, nelec, link_index=None, orbsym=None, wfnsym=0):
    if orbsym is None:
        return direct_spin1.contract_2e(eri, fcivec, norb, nelec, link_index)

    eri = ao2mo.restore(4, eri, norb)
    neleca, nelecb = _unpack_nelec(nelec)
    link_indexa, link_indexb = direct_spin1._unpack(norb, nelec, link_index)
    na, nlinka = link_indexa.shape[:2]
    nb, nlinkb = link_indexb.shape[:2]

    eri_irs, rank_eri, irrep_eri = reorder_eri(eri, norb, orbsym)
    strsa = cistring.gen_strings4orblist(range(norb), neleca)
    aidx, link_indexa = gen_str_irrep(strsa, orbsym, link_indexa, rank_eri, irrep_eri)
    nas = np.array([x.size for x in aidx], dtype=np.int32)
    if neleca == nelecb:
        bidx, link_indexb = aidx, link_indexa
        nbs = nas
    else:
        strsb = cistring.gen_strings4orblist(range(norb), nelecb)
        bidx, link_indexb = gen_str_irrep(strsb, orbsym, link_indexb, rank_eri, irrep_eri)
        nbs = np.array([x.size for x in bidx], dtype=np.int32)

    eri_ir_dims = np.array([x.shape[0] for x in eri_irs], dtype=np.int32)
    eri_irs = np.hstack([x.ravel() for x in eri_irs])

    wfnsym_in_d2h = wfnsym % 10
    orbsym_in_d2h = np.asarray(orbsym) % 10
    max_ir = orbsym_in_d2h.max()
    if max_ir >= 4:
        nirreps = 8
    elif max_ir >= 2:
        nirreps = 4
    elif max_ir >= 1:
        nirreps = 2
    else:
        nirreps = 1

    if fcivec.size == na * nb:
        fcivec_shape = fcivec.shape
        fcivec = fcivec.reshape((na,nb), order='C')
        ci0 = []
        for ir in range(nirreps):
            ma, mb = aidx[ir].size, bidx[wfnsym_in_d2h ^ ir].size
            ci0.append(np.zeros((ma, mb)))
            if ma * mb > 0:
                lib.take_2d(fcivec, aidx[ir], bidx[wfnsym_in_d2h ^ ir], out=ci0[ir])
        ci_size = np.array([x.size for x in ci0], dtype=np.int32)
        ci0 = np.hstack([x.ravel() for x in ci0])
    else:
        ci_size = []
        for ir in range(nirreps):
            ma, mb = aidx[ir].size, bidx[wfnsym_in_d2h ^ ir].size
            ci_size.append(ma * mb)
        ci_size = np.array(ci_size, dtype=np.int32)
        ci0 = fcivec
    ci1 = np.zeros_like(ci0)

    libfci.FCIcontract_2e_symm1(
        eri_irs.ctypes.data_as(ctypes.c_void_p),
        ci0.ctypes.data_as(ctypes.c_void_p),
        ci1.ctypes.data_as(ctypes.c_void_p),
        eri_ir_dims.ctypes.data_as(ctypes.c_void_p),
        ci_size.ctypes.data_as(ctypes.c_void_p),
        nas.ctypes.data_as(ctypes.c_void_p),
        nbs.ctypes.data_as(ctypes.c_void_p),
        link_indexa.ctypes.data_as(ctypes.c_void_p),
        link_indexb.ctypes.data_as(ctypes.c_void_p),
        ctypes.c_int(norb), ctypes.c_int(nlinka), ctypes.c_int(nlinkb),
        ctypes.c_int(nirreps), ctypes.c_int(wfnsym_in_d2h))

    if fcivec.size == na * nb:
        ci_loc = np.append(0, np.cumsum(ci_size))
        ci1new = np.zeros_like(fcivec)
        for ir in range(nirreps):
            if ci_size[ir] > 0:
                ma, mb = aidx[ir].size, bidx[wfnsym_in_d2h ^ ir].size
                buf = ci1[ci_loc[ir]:ci_loc[ir+1]].reshape(ma, mb)
                lib.takebak_2d(ci1new, buf, aidx[ir], bidx[wfnsym_in_d2h ^ ir])
        ci1 = ci1new.reshape(fcivec_shape)
    return ci1.view(direct_spin1.FCIvector)


def kernel(h1e, eri, norb, nelec, ci0=None, level_shift=1e-3, tol=1e-10,
           lindep=1e-14, max_cycle=50, max_space=12, nroots=1,
           davidson_only=False, pspace_size=400, orbsym=None, wfnsym=None,
           ecore=0, **kwargs):
    assert (len(orbsym) == norb)
    cis = FCISolver(None)
    cis.level_shift = level_shift
    cis.conv_tol = tol
    cis.lindep = lindep
    cis.max_cycle = max_cycle
    cis.max_space = max_space
    cis.nroots = nroots
    cis.davidson_only = davidson_only
    cis.pspace_size = pspace_size
    cis.orbsym = orbsym
    cis.wfnsym = wfnsym

    unknown = {}
    for k, v in kwargs.items():
        if not hasattr(cis, k):
            unknown[k] = v
        setattr(cis, k, v)
    if unknown:
        sys.stderr.write('Unknown keys %s for FCI kernel %s\n' %
                         (str(unknown.keys()), __name__))

    e, c = cis.kernel(h1e, eri, norb, nelec, ci0, ecore=ecore, **unknown)
    return e, c

make_rdm1 = direct_spin1.make_rdm1
make_rdm1s = direct_spin1.make_rdm1s
make_rdm12 = direct_spin1.make_rdm12

trans_rdm1s = direct_spin1.trans_rdm1s
trans_rdm1 = direct_spin1.trans_rdm1
trans_rdm12 = direct_spin1.trans_rdm12

def energy(h1e, eri, fcivec, norb, nelec, link_index=None, orbsym=None, wfnsym=0):
    h2e = direct_spin1.absorb_h1e(h1e, eri, norb, nelec) * .5
    ci1 = contract_2e(h2e, fcivec, norb, nelec, link_index, orbsym, wfnsym)
    return np.dot(fcivec.ravel(), ci1.ravel())

def _id_wfnsym(cisolver, norb, nelec, orbsym, wfnsym):
    '''Guess wfnsym or convert wfnsym to symmetry ID if it's a symmetry label'''
    gpname = getattr(cisolver.mol, 'groupname', None)
    if wfnsym is None:
        neleca, nelecb = _unpack_nelec(nelec)
        wfnsym = 0  # Ag, A1 or A
        for i in orbsym[nelecb:neleca]:
            wfnsym ^= i % 10
        if gpname in ('Dooh', 'Coov'):
            l = 0
            for i in orbsym[nelecb:neleca]:
                l += symm.basis.linearmole_irrep2momentum(i)
            wfnsym += (l//2) * 10
    elif isinstance(wfnsym, str):
        wfnsym = symm.irrep_name2id(gpname, wfnsym)
    return wfnsym

def _gen_strs_irrep(strs, orbsym):
    # % 10 to convert irrep_ids to irrep of D2h
    orbsym_in_d2h = np.asarray(orbsym) % 10
    irreps = np.zeros(len(strs), dtype=np.int32)
    if isinstance(strs, cistring.OIndexList):
        nocc = strs.shape[1]
        for i in range(nocc):
            irreps ^= orbsym_in_d2h[strs[:,i]]
    else:
        for i, ir in enumerate(orbsym_in_d2h):
            irreps[np.bitwise_and(strs, 1 << i) > 0] ^= ir
    return irreps

def _get_init_guess(airreps, birreps, nroots, hdiag, nelec, orbsym, wfnsym=0):
    neleca, nelecb = _unpack_nelec(nelec)
    na = len(airreps)
    nb = len(birreps)
    sym_allowed = airreps[:,None] == wfnsym ^ birreps
    if neleca == nelecb and na == nb:
        idx = np.arange(na)
        sym_allowed[idx[:,None] < idx] = False
    idx_a, idx_b = np.where(sym_allowed)

    hdiag = hdiag.reshape(na,nb)[idx_a,idx_b]
    if hdiag.size <= nroots:
        hdiag_indices = np.arange(hdiag.size)
    else:
        hdiag_indices = np.argpartition(hdiag, nroots-1)[:nroots]

    ci0 = []
    for k in hdiag_indices:
        addra, addrb = idx_a[k], idx_b[k]
        x = np.zeros((na, nb))
        x[addra,addrb] = 1
        ci0.append(x.ravel().view(direct_spin1.FCIvector))

    if len(ci0) == 0:
        raise lib.exceptions.WfnSymmetryError(
            f'Initial guess for symmetry {wfnsym} not found')
    return ci0

def get_init_guess(norb, nelec, nroots, hdiag, orbsym, wfnsym=0):
    neleca, nelecb = _unpack_nelec(nelec)
    na = cistring.num_strings(norb, neleca)
    nb = cistring.num_strings(norb, nelecb)
    if hdiag.size == na * nb:
        strsa = cistring.gen_strings4orblist(range(norb), neleca)
        airreps = birreps = _gen_strs_irrep(strsa, orbsym)
        if neleca != nelecb:
            strsb = cistring.gen_strings4orblist(range(norb), nelecb)
            birreps = _gen_strs_irrep(strsb, orbsym)
        return _get_init_guess(airreps, birreps, nroots, hdiag, nelec, orbsym, wfnsym)

    ci0 = []
    if hdiag.size <= nroots:
        hdiag_indices = np.arange(hdiag.size)
    else:
        hdiag_indices = np.argpartition(hdiag, nroots-1)[:nroots]
    for k in hdiag_indices:
        x = np.zeros_like(hdiag)
        x[k] = 1.
        ci0.append(x.ravel().view(direct_spin1.FCIvector))

    if len(ci0) == 0:
        raise lib.exceptions.WfnSymmetryError(
            f'Initial guess for symmetry {wfnsym} not found')
    return ci0

def _validate_degen_mapping(mapping, norb):
    '''Check if 2D irreps are properly paired'''
    mapping = np.asarray(mapping)
    return (mapping.max() < norb and
            # Must be self-conjugated
            numpy.array_equal(mapping[mapping], numpy.arange(norb)))

def get_init_guess_cyl_sym(norb, nelec, nroots, hdiag, orbsym, wfnsym=0):
    neleca, nelecb = _unpack_nelec(nelec)
    strsa = strsb = cistring.gen_strings4orblist(range(norb), neleca)
    airreps_d2h = birreps_d2h = _gen_strs_irrep(strsa, orbsym)
    a_ls = b_ls = _strs_angular_momentum(strsa, orbsym)
    if neleca != nelecb:
        strsb = cistring.gen_strings4orblist(range(norb), nelecb)
        birreps_d2h = _gen_strs_irrep(strsb, orbsym)
        b_ls = _strs_angular_momentum(strsb, orbsym)

    wfnsym_in_d2h = wfnsym % 10
    wfn_momentum = symm.basis.linearmole_irrep2momentum(wfnsym)
    na = len(strsa)
    nb = len(strsb)
    hdiag = hdiag.reshape(na,nb)
    degen_mapping = orbsym.degen_mapping
    ci0 = []
    iroot = 0
    wfn_ungerade = wfnsym_in_d2h >= 4
    a_ungerade = airreps_d2h >= 4
    b_ungerade = birreps_d2h >= 4
    sym_allowed = a_ungerade[:,None] == b_ungerade ^ wfn_ungerade
    # total angular momentum == wfn_momentum
    sym_allowed &= a_ls[:,None] == wfn_momentum - b_ls
    if neleca == nelecb and na == nb:
        idx = np.arange(na)
        sym_allowed[idx[:,None] < idx] = False
    idx_a, idx_b = np.where(sym_allowed)

    for k in hdiag[idx_a,idx_b].argsort():
        addra, addrb = idx_a[k], idx_b[k]
        ca = _cyl_sym_csf2civec(strsa, addra, orbsym, degen_mapping)
        cb = _cyl_sym_csf2civec(strsb, addrb, orbsym, degen_mapping)
        if wfnsym in (0, 1, 4, 5):
            addra1, sign_a = _sv_associated_det(strsa[addra], degen_mapping)
            addrb1, sign_b = _sv_associated_det(strsb[addrb], degen_mapping)
            if wfnsym in (1, 4) and addra == addra1 and addrb == addrb1:
                # Remove the A1 repr from initial guess.
                # The product of two E reprs can produce A1, A2 and another E repr.
                # addra == addra1 and addrb == addrb1  can be found in the A1 repr.
                # However, this may also incorrectly remove the A2 repr, see the
                # explanation in issue #2291.
                # In this case, the direct_spin1_cyl_sym solver can be used to
                # solve A2. See example mcscf/18-o2_spatial_spin_symmetry.py
                continue
            x = ca[:,None] * cb

            # If (E+) and (E-) are associated determinants
            # (E+)(E-') + (E-)(E+') => A1
            # (E+)(E-') - (E-)(E+') => A2
            if addra != addra1:
                ca = _cyl_sym_csf2civec(strsa, addra1, orbsym, degen_mapping)
            if addrb != addrb1:
                cb = _cyl_sym_csf2civec(strsb, addrb1, orbsym, degen_mapping)
            if wfnsym in (0, 5):  # A1g, A1u
                x += sign_a * sign_b * ca[:,None] * cb
                #assert (sign_a*sign_b==1 and x.imag==0) or (sign_a*sign_b==-1 and x.real==0)
            elif wfnsym in (1, 4):  # A2g, A2u
                x -= sign_a * sign_b * ca[:,None] * cb
                #assert (sign_a*sign_b==1 and x.real==0) or (sign_a*sign_b==-1 and x.imag==0)
            if np.linalg.norm(x.real) > 1e-6:
                x = x.real.copy()
            else:
                x = x.imag.copy()

        elif wfn_momentum > 0:
            x = ca.real[:,None] * cb.real
            x-= ca.imag[:,None] * cb.imag
        else:
            x = ca.imag[:,None] * cb.real
            x+= ca.real[:,None] * cb.imag

        norm = np.linalg.norm(x)
        if norm < 1e-3:
            continue
        x *= 1./norm
        ci0.append(x.ravel().view(direct_spin1.FCIvector))
        iroot += 1
        if iroot >= nroots:
            break

    if len(ci0) == 0:
        raise lib.exceptions.WfnSymmetryError(
            f'Initial guess for symmetry {wfnsym} not found')
    return ci0

def _cyl_sym_csf2civec(strs, addr, orbsym, degen_mapping):
    '''For orbital basis rotation from E(+/-) basis to Ex/Ey basis, mimic the CI
    transformation  addons.transform_ci(civec, (0, nelec), u)
    '''
    norb = orbsym.size
    one_particle_strs = np.asarray([1 << i for i in range(norb)])
    occ_masks = (strs[:,None] & one_particle_strs) != 0
    na = strs.size
    occ_idx_all_strs = np.where(occ_masks)[1].reshape(na,-1)

    u = _cyl_sym_orbital_rotation(orbsym, degen_mapping)
    ui = u[occ_masks[addr]].T.copy()
    minors = ui[occ_idx_all_strs]
    civec = np.linalg.det(minors)
    return civec

def _cyl_sym_orbital_rotation(orbsym, degen_mapping):
    '''Rotation to transform (E+)/(E-) basis to Ex/Ey basis
    |Ex/Ey> = |E(+/-)> * u
    '''
    norb = orbsym.size
    u = np.zeros((norb, norb), dtype=np.complex128)
    sqrth = .5**.5
    sqrthi = sqrth * 1j
    for i, j in enumerate(degen_mapping):
        if i == j:  # 1d irrep
            if orbsym[i] in (1, 4):  # A2g, A2u
                u[i,i] = 1j
            else:
                u[i,i] = 1
        elif orbsym[i] % 10 in (0, 2, 5, 7):  # Ex, E(+)
            u[j,j] = sqrthi
            u[i,j] = sqrthi
            u[j,i] = sqrth
            u[i,i] = -sqrth
    return u

def _sv_associated_det(ci_str, degen_mapping):
    '''Associated determinant for the sigma_v operation'''
    ci_str1 = 0
    nelec = 0
    sign = 1
    for i, j in enumerate(degen_mapping):
        if ci_str & (1 << i) > 0:
            if i > j and ci_str & (1 << j) > 0:
                # Ex, Ey orbitals swapped
                sign = -sign
            ci_str1 |= 1 << j
            nelec += 1
    return cistring.str2addr(degen_mapping.size, nelec, ci_str1), sign

def _strs_angular_momentum(strs, orbsym):
    # angular momentum for each orbital
    orb_l = (orbsym // 10) * 2
    e1_mask = np.isin(orbsym % 10, (2, 3, 6, 7))
    orb_l[e1_mask] += 1
    ey_mask = np.isin(orbsym % 10, (1, 3, 4, 6))
    orb_l[ey_mask] *= -1

    # total angular for each determinant (CSF)
    ls = np.zeros(len(strs), dtype=int)
    if isinstance(strs, cistring.OIndexList):
        nocc = strs.shape[1]
        for i in range(nocc):
            ls += orb_l[strs[:,i]]
    else:
        for i, l in enumerate(orb_l):
            ls[np.bitwise_and(strs, 1 << i) > 0] += l
    return ls

def reorder_eri(eri, norb, orbsym):
    if orbsym is None:
        return [eri], np.arange(norb), np.zeros(norb,dtype=np.int32)

    # % 10 to map irrep IDs of Dooh or Coov, etc. to irreps of D2h, C2v
    orbsym = np.asarray(orbsym) % 10

    # irrep of (ij| pair
    trilirrep = (orbsym[:,None] ^ orbsym)[np.tril_indices(norb)]
    # and the number of occurrences for each irrep
    dimirrep = np.asarray(np.bincount(trilirrep), dtype=np.int32)
    # we sort the irreps of (ij| pair, to group the pairs which have same irreps
    # "order" is irrep-id-sorted index. The (ij| paired is ordered that the
    # pair-id given by order[0] comes first in the sorted pair
    # "rank" is a sorted "order". Given nth (ij| pair, it returns the place(rank)
    # of the sorted pair
    old_eri_irrep = np.asarray(trilirrep, dtype=np.int32)
    rank_in_irrep = np.empty_like(old_eri_irrep)
    eri_irs = [np.zeros((0,0))] * TOTIRREPS
    for ir, nnorb in enumerate(dimirrep):
        idx = np.asarray(np.where(trilirrep == ir)[0], dtype=np.int32)
        rank_in_irrep[idx] = np.arange(nnorb, dtype=np.int32)
        eri_ir = lib.take_2d(eri, idx, idx)
        # Drop small integrals which may break symmetry?
        #eri_ir[abs(eri_ir) < 1e-13] = 0
        eri_irs[ir] = eri_ir
    return eri_irs, rank_in_irrep, old_eri_irrep

def argsort_strs_by_irrep(strs, orbsym):
    airreps = _gen_strs_irrep(strs, orbsym)
    aidx = [np.zeros(0,dtype=np.int32)] * TOTIRREPS
    for ir in range(TOTIRREPS):
        aidx[ir] = np.where(airreps == ir)[0]
    return aidx

def gen_str_irrep(strs, orbsym, link_index, rank_eri, irrep_eri):
    aidx = argsort_strs_by_irrep(strs, orbsym)
    na = len(strs)
    rank = np.zeros(na, dtype=np.int32)
    for idx in aidx:
        if idx.size > 0:
            rank[idx] = np.arange(idx.size, dtype=np.int32)

    link_index = link_index.copy()
    link_index[:,:,2] = rank[link_index[:,:,2]]
    link_index[:,:,1] = irrep_eri[link_index[:,:,0]]
    link_index[:,:,0] = rank_eri[link_index[:,:,0]]

    link_index = link_index.take(np.hstack(aidx), axis=0)
    return aidx, link_index

def _guess_wfnsym_cyl_sym(civec, strsa, strsb, orbsym):
    degen_mapping = orbsym.degen_mapping
    idx = abs(civec).argmax()
    na = strsa.size
    nb = strsb.size
    addra = idx // nb
    addrb = idx % nb
    addra1, sign_a = _sv_associated_det(strsa[addra], degen_mapping)
    addrb1, sign_b = _sv_associated_det(strsb[addrb], degen_mapping)
    addra, addra1 = min(addra,addra1), max(addra,addra1)
    addrb, addrb1 = min(addrb,addrb1), max(addrb,addrb1)
    ca = ca1 = _cyl_sym_csf2civec(strsa, addra, orbsym, degen_mapping)
    cb = cb1 = _cyl_sym_csf2civec(strsb, addrb, orbsym, degen_mapping)
    if addra != addra1:
        ca1 = _cyl_sym_csf2civec(strsa, addra1, orbsym, degen_mapping)
    if addrb != addrb1:
        cb1 = _cyl_sym_csf2civec(strsb, addrb1, orbsym, degen_mapping)
    ua = np.stack([ca, ca1])
    ub = np.stack([cb, cb1])
    # civec is in the Ex/Ey basis. Transform the largest coefficient to
    # (E+)/(E-) basis.
    c_max = ua.conj().dot(civec.reshape(na,nb)).dot(ub.conj().T)

    airreps_d2h = _gen_strs_irrep(strsa[[addra,addra1]], orbsym)
    birreps_d2h = _gen_strs_irrep(strsb[[addrb,addrb1]], orbsym)
    a_ls = _strs_angular_momentum(strsa[[addra,addra1]], orbsym)
    b_ls = _strs_angular_momentum(strsb[[addrb,addrb1]], orbsym)
    a_ungerade = airreps_d2h >= 4
    b_ungerade = birreps_d2h >= 4
    idx = abs(c_max).argmax()
    idx_a, idx_b = idx // 2, idx % 2
    wfn_ungerade = a_ungerade[idx_a] ^ b_ungerade[idx_b]
    wfn_momentum = a_ls[idx_a] + b_ls[idx_b]

    if wfn_momentum == 0:
        # For A1g and A1u, CI coefficient and its sigma_v associated one have
        # the same sign
        if (sign_a*sign_b * c_max[0,0].real * c_max[1,1].real > 1e-4 or
            sign_a*sign_b * c_max[0,0].imag * c_max[1,1].imag > 1e-4):  # A1
            if wfn_ungerade:
                wfnsym = 5
            else:
                wfnsym = 0
        elif (sign_a*sign_b * c_max[0,0].real * c_max[1,1].real < -1e-4 or
              sign_a*sign_b * c_max[0,0].imag * c_max[1,1].imag < -1e-4):  # A2
            # For A2g and A2u, CI coefficient and its sigma_v associated one
            # have opposite signs
            if wfn_ungerade:
                wfnsym = 4
            else:
                wfnsym = 1
        elif abs(c_max[0,1] - c_max[1,0]) < 1e-4: # Off-diagonal terms only
            # (E+)(E-') + (E-)(E+') => A1
            if wfn_ungerade:
                wfnsym = 5
            else:
                wfnsym = 0
        elif abs(c_max[0,1] + c_max[1,0]) < 1e-4:
            # (E+)(E-') - (E-)(E+') => A2
            if wfn_ungerade:
                wfnsym = 4
            else:
                wfnsym = 1
        else:
            raise RuntimeError('Symmetry broken wavefunction')

    elif wfn_momentum % 2 == 1:
        if abs(c_max[idx_a,idx_b].real) > 1e-6:  # Ex
            if wfn_ungerade:
                wfnsym = 7
            else:
                wfnsym = 2
        else:  # Ey
            if wfn_ungerade:
                wfnsym = 6
            else:
                wfnsym = 3
    else:
        if abs(c_max[idx_a,idx_b].real) > 1e-6:  # Ex
            if wfn_ungerade:
                wfnsym = 5
            else:
                wfnsym = 0
        else:  # Ey
            if wfn_ungerade:
                wfnsym = 4
            else:
                wfnsym = 1

    wfnsym += (abs(wfn_momentum) // 2) * 10
    return wfnsym

def guess_wfnsym(solver, norb, nelec, fcivec=None, orbsym=None, wfnsym=None, **kwargs):
    '''
    Guess point group symmetry of the FCI wavefunction.  If fcivec is
    given, the symmetry of fcivec is used.  Otherwise the symmetry is
    same to the HF determinant.
    '''
    if orbsym is None:
        orbsym = solver.orbsym

    verbose = kwargs.get('verbose', None)
    log = logger.new_logger(solver, verbose)
    neleca, nelecb = nelec = _unpack_nelec(nelec, solver.spin)

    groupname = getattr(solver.mol, 'groupname', None)
    if fcivec is None:
        # guess wfnsym if initial guess is not given
        wfnsym = _id_wfnsym(solver, norb, nelec, orbsym, wfnsym)
        log.debug('Guessing CI wfn symmetry = %s', wfnsym)

    elif wfnsym is None:
        if groupname in ('Dooh', 'Coov'):
            strsa = strsb = cistring.gen_strings4orblist(range(norb), neleca)
            if neleca != nelecb:
                strsb = cistring.gen_strings4orblist(range(norb), nelecb)
            if not isinstance(fcivec, np.ndarray) or fcivec.ndim > 2:
                fcivec = fcivec[0]
            wfnsym = _guess_wfnsym_cyl_sym(fcivec, strsa, strsb, orbsym)
        else:
            wfnsym = addons.guess_wfnsym(fcivec, norb, nelec, orbsym)
        log.debug('Guessing CI wfn symmetry = %s', wfnsym)

    else:
        # verify if the input wfnsym is consistent with the symmetry of fcivec
        strsa = strsb = cistring.gen_strings4orblist(range(norb), neleca)
        if neleca != nelecb:
            strsb = cistring.gen_strings4orblist(range(norb), nelecb)

        if groupname in ('Dooh', 'Coov'):
            if not isinstance(fcivec, np.ndarray) or fcivec.ndim > 2:
                fcivec = fcivec[0]
            wfnsym1 = _guess_wfnsym_cyl_sym(fcivec, strsa, strsb, orbsym)
            if wfnsym1 != _id_wfnsym(solver, norb, nelec, orbsym, wfnsym):
                raise lib.exceptions.WfnSymmetryError(
                    f'Input wfnsym {wfnsym} is not consistent with '
                    f'fcivec symmetry {wfnsym1}')
            wfnsym = wfnsym1
        else:
            na, nb = strsa.size, strsb.size
            orbsym_in_d2h = np.asarray(orbsym) % 10
            airreps = np.zeros(na, dtype=np.int32)
            birreps = np.zeros(nb, dtype=np.int32)
            for i, ir in enumerate(orbsym_in_d2h):
                airreps[np.bitwise_and(strsa, 1 << i) > 0] ^= ir
                birreps[np.bitwise_and(strsb, 1 << i) > 0] ^= ir

            wfnsym = _id_wfnsym(solver, norb, nelec, orbsym, wfnsym)
            groupname = getattr(solver.mol, 'groupname', None)
            mask = airreps[:,None] == (wfnsym % 10) ^ birreps

            if isinstance(fcivec, np.ndarray) and fcivec.ndim <= 2:
                fcivec = [fcivec]
            if all(abs(c.reshape(na, nb)[mask]).max() < 1e-5 for c in fcivec):
                raise lib.exceptions.WfnSymmetryError(
                    'Input wfnsym {wfnsym} is not consistent with fcivec coefficients')
    return wfnsym

def sym_allowed_indices(nelec, orbsym, wfnsym):
    '''Indices of symmetry allowed determinants for each irrep'''
    norb = orbsym.size
    neleca, nelecb = nelec
    strsa = strsb = cistring.gen_strings4orblist(range(norb), neleca)
    aidx = bidx = argsort_strs_by_irrep(strsa, orbsym)
    if neleca != nelecb:
        strsb = cistring.gen_strings4orblist(range(norb), nelecb)
        bidx = argsort_strs_by_irrep(strsb, orbsym)
    nb = len(strsb)
    wfnsym_in_d2h = wfnsym % 10
    ab_idx = [(aidx[ir][:,None] * nb + bidx[wfnsym_in_d2h ^ ir]).ravel()
              for ir in range(TOTIRREPS)]
    return ab_idx

class FCISolver(direct_spin1.FCISolver):

    _keys = {'wfnsym', 'sym_allowed_idx'}

    pspace_size = getattr(__config__, 'fci_direct_spin1_symm_FCI_pspace_size', 400)

    def __init__(self, mol=None, **kwargs):
        # wfnsym will be guessed based on initial guess if it is None
        self.wfnsym = None
        self.sym_allowed_idx = None
        direct_spin1.FCISolver.__init__(self, mol, **kwargs)

    def dump_flags(self, verbose=None):
        direct_spin1.FCISolver.dump_flags(self, verbose)
        log = logger.new_logger(self, verbose)
        if isinstance(self.wfnsym, str):
            log.info('Input CI wfn symmetry = %s', self.wfnsym)
        elif isinstance(self.wfnsym, (int, np.number)):
            groupname = getattr(self.mol, 'groupname', None)
            if groupname is not None:
                try:
                    log.info('Input CI wfn symmetry = %s',
                             symm.irrep_id2name(groupname, self.wfnsym))
                except KeyError:
                    raise RuntimeError('FCISolver cannot find mwfnsym Id %s in group %s. '
                                       'This might be caused by the projection from '
                                       'high-symmetry group to D2h symmetry.' %
                                       (self.wfnsym, groupname))
        else:
            log.info('CI wfn symmetry = %s', self.wfnsym)
        return self

    absorb_h1e = direct_spin1.FCISolver.absorb_h1e

    def make_hdiag(self, h1e, eri, norb, nelec, compress=False):
        nelec = _unpack_nelec(nelec, self.spin)
        hdiag = direct_spin1.make_hdiag(h1e, eri, norb, nelec)
        # TODO: hdiag should return symmetry allowed elements only. However,
        # get_init_guess_cyl_sym does not strictly follow the D2h (and subgroup)
        # symmetry treatments. The diagonal of entire Hamiltonian is required.
        if compress and self.sym_allowed_idx is not None:
            hdiag = hdiag.ravel()[np.hstack(self.sym_allowed_idx)]
        return hdiag

    def pspace(self, h1e, eri, norb, nelec, hdiag, np=400):
        nelec = _unpack_nelec(nelec, self.spin)
        na = cistring.num_strings(norb, nelec[0])
        nb = cistring.num_strings(norb, nelec[1])
        s_idx = numpy.hstack(self.sym_allowed_idx)
        if hdiag.size == s_idx.size:
            hdiag, hdiag0 = numpy.empty(na*nb), hdiag.ravel()
            hdiag[:] = 1e9
            hdiag[s_idx] = hdiag0
        elif not getattr(self.mol, 'groupname', None) in ('Dooh', 'Coov'):
            # Screen symmetry forbidden elements
            hdiag, hdiag0 = numpy.empty(na*nb), hdiag.ravel()
            hdiag[:] = 1e9
            hdiag[s_idx] = hdiag0[s_idx]

        np = min(np, hdiag.size)
        addr0, h = direct_spin1.pspace(h1e, eri, norb, nelec, hdiag, np)

        # mapping the address in (na,nb) civec to address in sym-allowed civec
        addr0_sym_allow = numpy.where(numpy.isin(addr0, s_idx))[0]
        addr0 = addr0[addr0_sym_allow]
        s_idx_allowed = numpy.where(numpy.isin(s_idx, addr0))[0]
        addr1 = s_idx[s_idx_allowed]
        new_idx = numpy.empty_like(s_idx_allowed)
        new_idx[addr0.argsort()] = addr1.argsort()
        addr = s_idx_allowed[new_idx]
        return addr, h[addr0_sym_allow[:,None],addr0_sym_allow]

    def contract_1e(self, f1e, fcivec, norb, nelec, link_index=None, **kwargs):
        nelec = direct_spin1._unpack_nelec(nelec)
        na = cistring.num_strings(norb, nelec[0])
        nb = cistring.num_strings(norb, nelec[1])
        if fcivec.size != na * nb:
            fcivec, ci0 = np.zeros(na*nb), fcivec
            fcivec[np.hstack(self.sym_allowed_idx)] = ci0
        return direct_spin1.contract_1e(f1e, fcivec, norb, nelec, link_index, **kwargs)

    def contract_2e(self, eri, fcivec, norb, nelec, link_index=None,
                    orbsym=None, wfnsym=None, **kwargs):
        if orbsym is None: orbsym = self.orbsym
        if wfnsym is None: wfnsym = self.wfnsym
        wfnsym = _id_wfnsym(self, norb, nelec, orbsym, wfnsym)
        nelec = _unpack_nelec(nelec, self.spin)
        return contract_2e(eri, fcivec, norb, nelec, link_index, orbsym, wfnsym, **kwargs)

    def contract_ss(self, fcivec, norb, nelec):
        nelec = direct_spin1._unpack_nelec(nelec)
        na = cistring.num_strings(norb, nelec[0])
        nb = cistring.num_strings(norb, nelec[1])
        if fcivec.size == na*nb:
            return contract_ss(fcivec, norb, nelec)

        fcivec, ci0 = np.zeros(na*nb), fcivec
        s_idx = np.hstack(self.sym_allowed_idx)
        fcivec[s_idx] = ci0
        ci1 = contract_ss(fcivec, norb, nelec)
        return ci1.ravel()[s_idx]

    def get_init_guess(self, norb, nelec, nroots, hdiag, orbsym=None, wfnsym=None):
        if orbsym is None: orbsym = self.orbsym
        if wfnsym is None:
            wfnsym = _id_wfnsym(self, norb, nelec, orbsym, self.wfnsym)
        s_idx = np.hstack(self.sym_allowed_idx)
        if getattr(self.mol, 'groupname', None) in ('Dooh', 'Coov'):
            ci0 = get_init_guess_cyl_sym(
                norb, nelec, nroots, hdiag, orbsym, wfnsym)
            return [x[s_idx] for x in ci0]
        else:
            return get_init_guess(norb, nelec, nroots, hdiag.ravel()[s_idx],
                                  orbsym, wfnsym)

    guess_wfnsym = guess_wfnsym

    def kernel(self, h1e, eri, norb, nelec, ci0=None,
               tol=None, lindep=None, max_cycle=None, max_space=None,
               nroots=None, davidson_only=None, pspace_size=None,
               orbsym=None, wfnsym=None, ecore=0, **kwargs):
        if nroots is None: nroots = self.nroots
        if orbsym is None: orbsym = self.orbsym
        if wfnsym is None: wfnsym = self.wfnsym
        if self.verbose >= logger.WARN:
            if 'verbose' not in kwargs:
                kwargs['verbose'] = self.verbose
            self.check_sanity()
        self.norb = norb
        self.nelec = nelec = _unpack_nelec(nelec, self.spin)
        link_index = direct_spin1._unpack(norb, nelec, None)

        if getattr(self.mol, 'groupname', None) in ('Dooh', 'Coov'):
            if not hasattr(orbsym, 'degen_mapping'):
                degen_mapping = map_degeneracy(h1e.diagonal(), orbsym)
                orbsym = lib.tag_array(orbsym, degen_mapping=degen_mapping)
            if davidson_only is None:
                davidson_only = True
            if not _validate_degen_mapping(orbsym.degen_mapping, norb):
                raise lib.exceptions.PointGroupSymmetryError(
                    'Incomplete 2D-irrep orbitals for cylindrical symmetry.\n'
                    f'orbsym = {orbsym}. '
                    f'Retry {self.__class__} with D2h subgroup symmetry.')

        wfnsym_ir = self.guess_wfnsym(norb, nelec, ci0, orbsym, wfnsym, **kwargs)
        self.sym_allowed_idx = sym_allowed_indices(nelec, orbsym, wfnsym_ir)
        s_idx = np.hstack(self.sym_allowed_idx)
        self.orbsym = orbsym
        logger.debug(self, 'Num symmetry allowed elements %d',
                     sum([x.size for x in self.sym_allowed_idx]))
        if s_idx.size == 0:
            raise lib.exceptions.WfnSymmetryError(
                f'Symmetry allowed determinants not found for wfnsym {wfnsym}')

        if wfnsym_ir > 7:
            # Symmetry broken for Dooh and Coov groups is often observed.
            # A larger max_space is helpful to reduce the error. Also it is
            # hard to converge to high precision.
            if max_space is None and self.max_space == FCISolver.max_space:
                max_space = 20 + 7 * nroots
            if tol is None and self.conv_tol == FCISolver.conv_tol:
                tol = 1e-7

        if ci0 is None and getattr(self.mol, 'groupname', None) in ('Dooh', 'Coov'):
            # self.hdiag returns stripped H_diag (for D2h symmetry).
            # Different convention of symmetry representations were used in
            # get_init_guess_cyl_sym (which follows direct_spin1_cyl_sym.py).
            # Some symmetry forbidden elements for D2h are needed in
            # get_init_guess_cyl_sym function. Thus the entire hdiag is computed.
            hdiag = self.make_hdiag(h1e, eri, norb, nelec, compress=False)
            ci0 = self.get_init_guess(norb, nelec, nroots, hdiag, orbsym, wfnsym_ir)

        with lib.temporary_env(self, wfnsym=wfnsym_ir):
            e, c = direct_spin1.kernel_ms1(
                self, h1e, eri, norb, nelec, ci0, link_index, tol, lindep, max_cycle,
                max_space, nroots, davidson_only, pspace_size, ecore=ecore, **kwargs)

        na = link_index[0].shape[0]
        nb = link_index[1].shape[0]
        if nroots > 1:
            c, c_raw = [], c
            for vec in c_raw:
                c1 = np.zeros(na*nb)
                c1[s_idx] = vec.T
                c.append(c1.reshape(na, nb).view(direct_spin1.FCIvector))
        else:
            c1 = np.zeros(na*nb)
            c1[s_idx] = c
            c = c1.reshape(na, nb).view(direct_spin1.FCIvector)

        self.eci, self.ci = e, c
        return e, c

FCI = FCISolver



#!/usr/bin/env python
# Copyright 2021-2024 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Xing Zhang <zhangxing.nju@gmail.com>
#

import numpy as np
from pyscf import __config__
from pyscf import lib
from pyscf.lib import logger
from pyscf.gto import moleintor
from pyscf.pbc import tools
from pyscf.pbc.lib.kpts_helper import gamma_point
from pyscf.pbc.lib.kpts import KPoints
from pyscf.pbc.df import fft
from pyscf.pbc import tools as pbctools
from pyscf.pbc.df.df_jk import (
    _format_dms,
    _format_kpts_band,
    _format_jks,
)
from pyscf.pbc.dft.multigrid.pp import (
    _get_vpplocG_part1,
    _get_pp_without_erf,
    vpploc_part1_nuc_grad,
    get_vpploc_part1_ip1,
)
from pyscf.pbc.dft.multigrid.utils import (
    _take_4d,
    _take_5d,
    _takebak_4d,
    _takebak_5d,
)
from pyscf.pbc.dft.multigrid.multigrid import MultiGridNumInt as MultiGridNumInt_v1
from pyscf.pbc.dft.multigrid import _backend_c as backend

NTASKS = getattr(__config__, 'pbc_dft_multigrid_ntasks', 4)
KE_RATIO = getattr(__config__, 'pbc_dft_multigrid_ke_ratio', 3.0)
REL_CUTOFF = getattr(__config__, 'pbc_dft_multigrid_rel_cutoff', 20.0)
GGA_METHOD = getattr(__config__, 'pbc_dft_multigrid_gga_method', 'FFT')

EXTRA_PREC = getattr(__config__, 'pbc_gto_eval_gto_extra_precision', 1e-2)
RHOG_HIGH_ORDER = getattr(__config__, 'pbc_dft_multigrid_rhog_high_order', False)
PTR_EXPDROP = 16


def multi_grids_tasks(cell, ke_cutoff=None, hermi=0,
                      ntasks=NTASKS, ke_ratio=KE_RATIO, rel_cutoff=REL_CUTOFF):
    if ke_cutoff is None:
        ke_cutoff = cell.ke_cutoff
    a = cell.lattice_vectors()
    if ke_cutoff is None:
        ke_cutoff = pbctools.mesh_to_cutoff(a, cell.mesh).max()
    ke1 = ke_cutoff
    cutoff = [ke1,]
    for i in range(ntasks-1):
        ke1 /= ke_ratio
        cutoff.append(ke1)
    cutoff.reverse()
    mesh = []
    for ke in cutoff[:-1]:
        mesh.append(tools.cutoff_to_mesh(a, ke))
    mesh.append(cell.mesh)
    logger.info(cell, 'ke_cutoff for multigrid tasks:\n%s', cutoff)
    logger.info(cell, 'meshes for multigrid tasks:\n%s', mesh)
    gridlevel_info = backend.GridLevel_Info(cutoff, rel_cutoff, mesh)
    task_list = backend.TaskList(cell, gridlevel_info, hermi=hermi)
    return task_list


def _update_task_list(mydf, hermi=0, ntasks=None, ke_ratio=None, rel_cutoff=None):
    '''Update :attr:`task_list` if necessary.
    '''
    cell = mydf.cell
    if ntasks is None:
        ntasks = mydf.ntasks
    if ke_ratio is None:
        ke_ratio = mydf.ke_ratio
    if rel_cutoff is None:
        rel_cutoff = mydf.rel_cutoff

    need_update = False
    task_list = getattr(mydf, 'task_list', None)
    if task_list is None:
        need_update = True
    else:
        ke_cutoff = cell.ke_cutoff
        if ke_cutoff is None:
            a = cell.lattice_vectors()
            ke_cutoff = pbctools.mesh_to_cutoff(a, cell.mesh).max()
        hermi_orig = task_list.hermi
        nlevels = task_list.nlevels
        rel_cutoff_orig = task_list.gridlevel_info.rel_cutoff
        ke_cutoff_orig = task_list.gridlevel_info.cutoff[-1]
        if (hermi_orig > hermi or
                nlevels != ntasks or
                abs(rel_cutoff_orig-rel_cutoff) > 1e-12 or
                abs(ke_cutoff_orig - ke_cutoff) > 1e-12):
            need_update = True
            logger.debug(mydf, 'Hermiticity or cutoffs changed; will update the task list!')

    if need_update:
        task_list = multi_grids_tasks(cell, hermi=hermi, ntasks=ntasks,
                                      ke_ratio=ke_ratio, rel_cutoff=rel_cutoff)
        mydf.task_list = task_list
    return task_list


def eval_rho(cell, dm, task_list, shls_slice=None, hermi=0, xctype='LDA', kpts=None,
             dimension=None, cell1=None, shls_slice1=None,
             a=None, ignore_imag=False):
    '''Collocate density (and gradients) on the real-space grid.

    The two sets of Gaussian basis functions can be different.

    Returns:
        rho: `RS_Grid` object
            Densities on real space multigrids.
    '''
    cell0 = cell
    shls_slice0 = shls_slice
    if cell1 is None:
        cell1 = cell0

    #TODO mixture of cartesian and spherical bases
    assert cell0.cart == cell1.cart

    ish_atm = cell0._atm
    ish_bas = cell0._bas
    ish_env = cell0._env
    ish_env[PTR_EXPDROP] = cell0.precision * EXTRA_PREC

    if cell1 is cell0:
        jsh_atm = ish_atm
        jsh_bas = ish_bas
        jsh_env = ish_env
    else:
        jsh_atm = cell1._atm
        jsh_bas = cell1._bas
        jsh_env = cell1._env
        jsh_env[PTR_EXPDROP] = cell1.precision * EXTRA_PREC

    if shls_slice0 is None:
        shls_slice0 = (0, cell0.nbas)
    i0, i1 = shls_slice0
    if shls_slice1 is None:
        shls_slice1 = shls_slice0
    j0, j1 = shls_slice1

    if hermi == 1:
        assert cell1 is cell0
        assert i0 == j0 and i1 == j1

    key0 = 'cart' if cell0.cart else 'sph'
    ao_loc0 = moleintor.make_loc(ish_bas, key0)
    naoi = ao_loc0[i1] - ao_loc0[i0]
    if hermi == 1:
        ao_loc1 = ao_loc0
    else:
        key1 = 'cart' if cell1.cart else 'sph'
        ao_loc1 = moleintor.make_loc(jsh_bas, key1)
    naoj = ao_loc1[j1] - ao_loc1[j0]

    dm = np.asarray(dm)
    assert dm.shape[-2:] == (naoi, naoj)

    if dimension is None:
        dimension = cell0.dimension
    assert dimension == getattr(cell1, "dimension", None)

    if dimension == 0 or kpts is None or gamma_point(kpts):
        dm = dm.reshape(-1,1,naoi,naoj)
    else:
        raise NotImplementedError
    n_dm = dm.shape[0]

    if a is None:
        a = cell0.lattice_vectors()
        if cell1 is not cell:
            a1 = cell1.lattice_vectors()
            if abs(a-a1).max() > 1e-12:
                raise RuntimeError('The two cell objects must have the same lattice vectors.')
    b = np.linalg.inv(a.T)

    rho = []
    for i, dm_i in enumerate(dm):
        rho_i = backend.grid_collocate(
                    xctype, dm_i,
                    task_list, hermi,
                    (i0, i1, j0, j1),
                    ao_loc0, ao_loc1, dimension,
                    a, b,
                    ish_atm, ish_bas, ish_env,
                    jsh_atm, jsh_bas, jsh_env,
                    cell0.cart)
        rho.append(rho_i)

    if n_dm == 1:
        rho = rho[0]
    return rho


def _eval_rhoG(mydf, dm_kpts, hermi=1, kpts=np.zeros((1,3)), deriv=0,
               rhog_high_order=RHOG_HIGH_ORDER):
    if deriv >= 2:
        raise NotImplementedError
    cell = mydf.cell

    dm_kpts = np.asarray(dm_kpts)
    dms = _format_dms(dm_kpts, kpts)
    nset, nkpts, nao = dms.shape[:3]

    task_list = _update_task_list(mydf, hermi=hermi, ntasks=mydf.ntasks,
                                  ke_ratio=mydf.ke_ratio, rel_cutoff=mydf.rel_cutoff)

    gga_high_order = False
    if deriv == 0:
        xctype = 'LDA'
        rhodim = 1
    elif deriv == 1:
        if rhog_high_order:
            raise NotImplementedError
        else:  # approximate high order derivatives in reciprocal space
            gga_high_order = True
            xctype = 'LDA'
            rhodim = 1
            deriv = 0
        assert hermi == 1 or gamma_point(kpts)

    ignore_imag = (hermi == 1)

    nx, ny, nz = mydf.mesh
    mem_avail = mydf.max_memory - lib.current_memory()[0]
    mem_needed = rhodim * nx * ny * nz * lib.num_threads() * 8 / 1e6
    if mem_needed > mem_avail:
        logger.warn(mydf, f'At least {mem_needed} MB of memory is needed for eval_rho. '
                    f'Currently {mem_avail} MB of memory is available.')
    rs_rho = eval_rho(cell, dms, task_list, hermi=hermi, xctype=xctype, kpts=kpts,
                      ignore_imag=ignore_imag)

    rhoG = np.zeros((nset*rhodim,nx,ny,nz), dtype=np.complex128)
    for ilevel, mesh in enumerate(task_list.gridlevel_info.mesh):
        ngrids = np.prod(mesh)
        if nset > 1:
            rho = []
            for i in range(nset):
                rho.append(rs_rho[i][ilevel])
            rho = np.asarray(rho)
        else:
            rho = rs_rho[ilevel]

        weight = 1./nkpts * cell.vol/ngrids
        rho_freq = tools.fft(rho.reshape(nset*rhodim, -1), mesh)
        rho = None
        rho_freq *= weight
        gx = np.fft.fftfreq(mesh[0], 1./mesh[0]).astype(np.int32)
        gy = np.fft.fftfreq(mesh[1], 1./mesh[1]).astype(np.int32)
        gz = np.fft.fftfreq(mesh[2], 1./mesh[2]).astype(np.int32)
        _takebak_4d(rhoG, rho_freq.reshape((-1,) + tuple(mesh)), (None, gx, gy, gz))
        rho_freq = None

    rs_rho = None

    rhoG = rhoG.reshape(nset,rhodim,-1)
    if gga_high_order:
        Gv = cell.get_Gv(mydf.mesh)
        #:rhoG1 = np.einsum('np,px->nxp', rhoG[:,0], 1j*Gv)
        rhoG1 = backend.gradient_gs(rhoG[:,0], Gv)
        rhoG = np.concatenate([rhoG, rhoG1], axis=1)
        Gv = rhoG1 = None
    return rhoG


def eval_mat(cell, weights, task_list, shls_slice=None, comp=1, hermi=0, deriv=0,
             xctype='LDA', kpts=None, grid_level=None, dimension=None, mesh=None,
             cell1=None, shls_slice1=None, a=None):
    if deriv == 1:
        assert comp == 3
        assert hermi == 0
    elif deriv > 1:
        raise NotImplementedError

    cell0 = cell
    shls_slice0 = shls_slice
    if cell1 is None:
        cell1 = cell0

    if mesh is None:
        mesh = cell0.mesh

    #TODO mixture of cartesian and spherical bases
    assert cell0.cart == cell1.cart

    ish_atm = cell0._atm
    ish_bas = cell0._bas
    ish_env = cell0._env
    ish_env[PTR_EXPDROP] = cell0.precision * EXTRA_PREC

    if cell1 is cell0:
        jsh_atm = ish_atm
        jsh_bas = ish_bas
        jsh_env = ish_env
    else:
        jsh_atm = cell1._atm
        jsh_bas = cell1._bas
        jsh_env = cell1._env
        jsh_env[PTR_EXPDROP] = cell1.precision * EXTRA_PREC

    if shls_slice0 is None:
        shls_slice0 = (0, cell0.nbas)
    i0, i1 = shls_slice0
    if shls_slice1 is None:
        shls_slice1 = (0, cell1.nbas)
    j0, j1 = shls_slice1

    if hermi == 1:
        assert cell1 is cell0
        assert i0 == j0 and i1 == j1

    key0 = 'cart' if cell0.cart else 'sph'
    ao_loc0 = moleintor.make_loc(ish_bas, key0)
    if hermi == 1:
        ao_loc1 = ao_loc0
    else:
        key1 = 'cart' if cell1.cart else 'sph'
        ao_loc1 = moleintor.make_loc(jsh_bas, key1)

    if dimension is None:
        dimension = cell0.dimension
    assert dimension == getattr(cell1, "dimension", None)

    weights = np.asarray(weights)
    if dimension == 0 or kpts is None or gamma_point(kpts):
        assert weights.dtype == np.double
    else:
        raise NotImplementedError

    if a is None:
        a = cell0.lattice_vectors()
        if cell1 is not cell:
            a1 = cell1.lattice_vectors()
            if abs(a-a1).max() > 1e-12:
                raise RuntimeError('The two cell objects must have the same lattice vectors.')
    b = np.linalg.inv(a.T)

    xctype = xctype.upper()
    n_mat = None
    if xctype == 'LDA':
        if weights.ndim == 1:
            weights = weights.reshape(-1, np.prod(mesh))
        else:
            n_mat = weights.shape[0]
    elif xctype == 'GGA':
        if weights.ndim == 2:
            weights = weights.reshape(-1, 4, np.prod(mesh))
        else:
            n_mat = weights.shape[0]
    else:
        raise NotImplementedError

    out = []
    for wv in weights:
        mat = backend.grid_integrate(
                xctype, wv,
                task_list, comp, hermi, grid_level,
                (i0, i1, j0, j1),
                ao_loc0, ao_loc1, dimension,
                a, b,
                ish_atm, ish_bas, ish_env,
                jsh_atm, jsh_bas, jsh_env,
                cell0.cart)
        out.append(mat)

    if n_mat is None:
        out = out[0]
    return out


def _get_j_pass2(mydf, vG, kpts=np.zeros((1,3)), hermi=1, verbose=None):
    cell = mydf.cell
    nkpts = len(kpts)
    nao = cell.nao_nr()
    nx, ny, nz = mydf.mesh
    vG = vG.reshape(-1,nx,ny,nz)
    nset = vG.shape[0]

    task_list = _update_task_list(mydf, hermi=hermi, ntasks=mydf.ntasks,
                                  ke_ratio=mydf.ke_ratio, rel_cutoff=mydf.rel_cutoff)

    if gamma_point(kpts):
        vj_kpts = np.zeros((nset,nkpts,nao,nao))
    else:
        raise NotImplementedError

    nlevels = task_list.nlevels
    meshes = task_list.gridlevel_info.mesh
    for ilevel in range(nlevels):
        mesh = meshes[ilevel]
        ngrids = np.prod(mesh)

        gx = np.fft.fftfreq(mesh[0], 1./mesh[0]).astype(np.int32)
        gy = np.fft.fftfreq(mesh[1], 1./mesh[1]).astype(np.int32)
        gz = np.fft.fftfreq(mesh[2], 1./mesh[2]).astype(np.int32)
        sub_vG = _take_4d(vG, (None, gx, gy, gz)).reshape(nset,ngrids)

        v_rs = tools.ifft(sub_vG, mesh).reshape(nset,ngrids)
        vR = np.asarray(v_rs.real, order='C')
        mat = eval_mat(cell, vR, task_list, comp=1, hermi=hermi,
                       xctype='LDA', kpts=kpts, grid_level=ilevel, mesh=mesh)
        vj_kpts += np.asarray(mat).reshape(nset,-1,nao,nao)

    if nset == 1:
        vj_kpts = vj_kpts[0]
    return vj_kpts


def _get_j_pass2_ip1(mydf, vG, kpts=np.zeros((1,3)), hermi=0, deriv=1, verbose=None):
    if deriv == 1:
        comp = 3
        assert hermi == 0
    else:
        raise NotImplementedError

    cell = mydf.cell
    nkpts = len(kpts)
    nao = cell.nao_nr()
    nx, ny, nz = mydf.mesh
    vG = vG.reshape(-1,nx,ny,nz)
    nset = vG.shape[0]

    task_list = _update_task_list(mydf, hermi=hermi, ntasks=mydf.ntasks,
                                  ke_ratio=mydf.ke_ratio, rel_cutoff=mydf.rel_cutoff)

    if gamma_point(kpts):
        vj_kpts = np.zeros((nset,nkpts,comp,nao,nao))
    else:
        raise NotImplementedError

    nlevels = task_list.nlevels
    meshes = task_list.gridlevel_info.mesh
    for ilevel in range(nlevels):
        mesh = meshes[ilevel]
        ngrids = np.prod(mesh)

        gx = np.fft.fftfreq(mesh[0], 1./mesh[0]).astype(np.int32)
        gy = np.fft.fftfreq(mesh[1], 1./mesh[1]).astype(np.int32)
        gz = np.fft.fftfreq(mesh[2], 1./mesh[2]).astype(np.int32)
        sub_vG = _take_4d(vG, (None, gx, gy, gz)).reshape(nset,ngrids)

        v_rs = tools.ifft(sub_vG, mesh).reshape(nset,ngrids)
        vR = np.asarray(v_rs.real, order='C')
        mat = eval_mat(cell, vR, task_list, comp=comp, hermi=hermi, deriv=deriv,
                       xctype='LDA', kpts=kpts, grid_level=ilevel, mesh=mesh)
        mat = np.asarray(mat).reshape(nset,-1,comp,nao,nao)
        vj_kpts += mat

    if nset == 1:
        vj_kpts = vj_kpts[0]
    return vj_kpts


def _get_gga_pass2(mydf, vG, kpts=np.zeros((1,3)), hermi=1, verbose=None):
    cell = mydf.cell
    nkpts = len(kpts)
    nao = cell.nao_nr()
    nx, ny, nz = mydf.mesh
    vG = vG.reshape(-1,4,nx,ny,nz)
    nset = vG.shape[0]

    task_list = _update_task_list(mydf, hermi=hermi, ntasks=mydf.ntasks,
                                  ke_ratio=mydf.ke_ratio, rel_cutoff=mydf.rel_cutoff)

    if gamma_point(kpts):
        veff = np.zeros((nset,nkpts,nao,nao))
    else:
        raise NotImplementedError

    nlevels = task_list.nlevels
    meshes = task_list.gridlevel_info.mesh
    for ilevel in range(nlevels):
        mesh = meshes[ilevel]
        ngrids = np.prod(mesh)

        gx = np.fft.fftfreq(mesh[0], 1./mesh[0]).astype(np.int32)
        gy = np.fft.fftfreq(mesh[1], 1./mesh[1]).astype(np.int32)
        gz = np.fft.fftfreq(mesh[2], 1./mesh[2]).astype(np.int32)
        sub_vG = _take_5d(vG, (None, None, gx, gy, gz)).reshape(-1,ngrids)
        wv = tools.ifft(sub_vG, mesh).reshape(nset,4,ngrids)
        wv = np.asarray(wv.real, order='C')
        mat = eval_mat(cell, wv, task_list, comp=1, hermi=hermi,
                       xctype='GGA', kpts=kpts, grid_level=ilevel, mesh=mesh)
        mat = np.asarray(mat).reshape(nset,-1,nao,nao)
        veff += mat

    if nset == 1:
        veff = veff[0]
    return veff


def _get_gga_pass2_ip1(mydf, vG, kpts=np.zeros((1,3)), hermi=0, deriv=1, verbose=None):
    if deriv == 1:
        comp = 3
        assert hermi == 0
    else:
        raise NotImplementedError

    cell = mydf.cell
    nkpts = len(kpts)
    nao = cell.nao_nr()
    nx, ny, nz = mydf.mesh
    vG = vG.reshape(-1,4,nx,ny,nz)
    nset = vG.shape[0]

    task_list = _update_task_list(mydf, hermi=hermi, ntasks=mydf.ntasks,
                                  ke_ratio=mydf.ke_ratio, rel_cutoff=mydf.rel_cutoff)

    if gamma_point(kpts):
        vj_kpts = np.zeros((nset,nkpts,comp,nao,nao))
    else:
        raise NotImplementedError

    for ilevel, mesh in enumerate(task_list.gridlevel_info.mesh):
        ngrids = np.prod(mesh)

        gx = np.fft.fftfreq(mesh[0], 1./mesh[0]).astype(np.int32)
        gy = np.fft.fftfreq(mesh[1], 1./mesh[1]).astype(np.int32)
        gz = np.fft.fftfreq(mesh[2], 1./mesh[2]).astype(np.int32)
        sub_vG = _take_5d(vG, (None, None, gx, gy, gz)).reshape(-1,ngrids)

        v_rs = tools.ifft(sub_vG, mesh).reshape(nset,4,ngrids)
        vR = np.asarray(v_rs.real, order='C')
        mat = eval_mat(cell, vR, task_list, comp=comp, hermi=hermi, deriv=deriv,
                       xctype='GGA', kpts=kpts, grid_level=ilevel, mesh=mesh)
        vj_kpts += np.asarray(mat).reshape(nset,-1,comp,nao,nao)

    if nset == 1:
        vj_kpts = vj_kpts[0]
    return vj_kpts


def nr_rks(mydf, xc_code, dm_kpts, hermi=1, kpts=None,
           kpts_band=None, with_j=False, return_j=False, verbose=None):
    '''Compute the XC energy and RKS XC matrix using the multigrid algorithm.

    See also `multigrid.nr_rks`.
    '''
    if kpts is None:
        kpts = np.zeros((1, 3))
    elif isinstance(kpts, KPoints):
        if kpts.kpts.size > 3: # multiple k points
            dm_kpts = kpts.transform_dm(dm_kpts)
        kpts = kpts.kpts
    log = logger.new_logger(mydf, verbose)
    cell = mydf.cell
    dm_kpts = np.asarray(dm_kpts, order='C')
    dms = _format_dms(dm_kpts, kpts)
    nset, nkpts, nao = dms.shape[:3]
    assert nset == 1
    kpts_band, input_band = _format_kpts_band(kpts_band, kpts), kpts_band

    ni = mydf
    xctype = ni._xc_type(xc_code)
    if xctype in (None, 'LDA', 'HF'):
        deriv = 0
    elif xctype == 'GGA':
        deriv = 1
    else:
        raise NotImplementedError
    rhoG = _eval_rhoG(mydf, dm_kpts, hermi, kpts, deriv)

    mesh = mydf.mesh
    ngrids = np.prod(mesh)

    coulG = tools.get_coulG(cell, mesh=mesh)

    #:vG = np.einsum('ng,g->ng', rhoG[:,0], coulG)
    vG = np.multiply(rhoG[0,0], coulG)
    coulG = None

    if mydf.vpplocG_part1 is not None:
        vG += mydf.vpplocG_part1 * 2

    #:ecoul = .5 * np.einsum('ng,ng->n', rhoG[:,0].real, vG.real)
    #:ecoul+= .5 * np.einsum('ng,ng->n', rhoG[:,0].imag, vG.imag)
    ecoul = .5 * np.vdot(rhoG[0,0], vG).real

    ecoul /= cell.vol
    log.debug('Multigrid Coulomb energy %s', ecoul)

    if mydf.vpplocG_part1 is not None:
        vG -= mydf.vpplocG_part1

    weight = cell.vol / ngrids
    # *(1./weight) because rhoR is scaled by weight in _eval_rhoG.  When
    # computing rhoR with IFFT, the weight factor is not needed.
    rhoR = tools.ifft(rhoG.reshape(-1,ngrids), mesh).real * (1./weight)
    rhoR = rhoR.reshape(-1,ngrids)
    exc, vxc = ni.eval_xc_eff(xc_code, rhoR, deriv=1, xctype=xctype)[:2]
    wv_freq = tools.fft(vxc, mesh).reshape(-1,ngrids)
    if xctype == 'GGA' and GGA_METHOD.upper() == 'FFT':
        #:wv_freq = (wv_freq[0] - 1j * np.einsum('px,xp->p', Gv, wv_freq[1:4])) * weight
        Gv = cell.get_Gv(ni.mesh)
        wv_freq = backend.get_gga_vrho_gs(wv_freq[:1], wv_freq[1:4], Gv, weight, ngrids, 1.)
    else:
        wv_freq *= weight

    nelec = np.sum(rhoR[0]) * weight
    excsum = np.dot(rhoR[0], exc) * weight
    exc = vxc = None
    rhoR = rhoG = None
    log.debug('Multigrid exc %s  nelec %s', excsum, nelec)

    if with_j:
        wv_freq[0] += vG

    kpts_band, input_band = _format_kpts_band(kpts_band, kpts), kpts_band
    if xctype in (None, 'LDA', 'HF'):
        veff = _get_j_pass2(mydf, wv_freq, kpts_band, verbose=log)
    elif xctype == 'GGA':
        if GGA_METHOD.upper() == 'FFT':
            veff = _get_j_pass2(mydf, wv_freq, kpts_band, verbose=log)
        else:
            veff = _get_gga_pass2(mydf, wv_freq, kpts_band, hermi=hermi, verbose=log)
    else:
        raise NotImplementedError
    wv_freq = None
    veff = _format_jks(veff, dm_kpts, input_band, kpts)

    if return_j:
        vj = _get_j_pass2(mydf, vG, kpts_band, verbose=log)
        vj = _format_jks(veff, dm_kpts, input_band, kpts)
    else:
        vj = None
    vG = None

    veff = lib.tag_array(veff, ecoul=ecoul, exc=excsum, vj=vj, vk=None)
    return nelec, excsum, veff

def nr_uks(mydf, xc_code, dm_kpts, hermi=1, kpts=None,
           kpts_band=None, with_j=False, return_j=False, verbose=None):
    if kpts is None:
        kpts = np.zeros((1, 3))
    elif isinstance(kpts, KPoints):
        if kpts.kpts.size > 3: # multiple k points
            dm_kpts = kpts.transform_dm(dm_kpts)
        kpts = kpts.kpts
    log = logger.new_logger(mydf, verbose)
    cell = mydf.cell
    dm_kpts = np.asarray(dm_kpts, order='C')
    dms = _format_dms(dm_kpts, kpts)
    nset, nkpts, nao = dms.shape[:3]
    nset //= 2
    assert nset == 1
    kpts_band, input_band = _format_kpts_band(kpts_band, kpts), kpts_band

    mesh = mydf.mesh
    ngrids = np.prod(mesh)
    ni = mydf
    xctype = ni._xc_type(xc_code)
    if xctype in (None, 'LDA', 'HF'):
        deriv = 0
    elif xctype == 'GGA':
        deriv = 1
    else:
        raise NotImplementedError

    rhoG = _eval_rhoG(mydf, dm_kpts, hermi, kpts, deriv)
    rhoG = rhoG.reshape(2,-1,ngrids)
    rhoG_sf = rhoG[0,0] + rhoG[1,0]

    coulG = tools.get_coulG(cell, mesh=mesh)
    vG = rhoG_sf * coulG
    coulG = None

    if mydf.vpplocG_part1 is not None:
        vG += mydf.vpplocG_part1 * 2

    ecoul = .5 * np.vdot(rhoG_sf, vG).real

    ecoul /= cell.vol
    log.debug('Multigrid Coulomb energy %s', ecoul)

    if mydf.vpplocG_part1 is not None:
        vG -= mydf.vpplocG_part1

    weight = cell.vol / ngrids
    # *(1./weight) because rhoR is scaled by weight in _eval_rhoG.  When
    # computing rhoR with IFFT, the weight factor is not needed.
    rhoR = tools.ifft(rhoG.reshape(-1,ngrids), mesh).real * (1./weight)
    rhoR = rhoR.reshape(2,-1,ngrids)
    exc, vxc = ni.eval_xc_eff(xc_code, rhoR, deriv=1, xctype=xctype)[:2]
    wv_freq = tools.fft(vxc.reshape(-1,ngrids), mesh).reshape(2,-1,ngrids)
    if xctype == 'GGA' and GGA_METHOD.upper() == 'FFT':
        #:wv_freq = (wv_freq[:,0] - 1j * np.einsum('px,nxp->np', Gv, wv_freq[:,1:4])) * weight
        Gv = cell.get_Gv(ni.mesh)
        backend.get_gga_vrho_gs(wv_freq[0,:1], wv_freq[0,1:4], Gv, weight, ngrids, 1.)
        backend.get_gga_vrho_gs(wv_freq[1,:1], wv_freq[1,1:4], Gv, weight, ngrids, 1.)
        wv_freq = wv_freq[:,:1]
    else:
        wv_freq *= weight

    nelec = rhoR[:,0].sum(axis=1) * weight
    excsum = (rhoR[0,0] + rhoR[1,0]).dot(exc) * weight
    exc = vxc = None

    rhoR = rhoG = None
    log.debug('Multigrid exc %s  nelec %s', excsum, nelec)

    if with_j:
        wv_freq[:,0] += vG

    kpts_band, input_band = _format_kpts_band(kpts_band, kpts), kpts_band
    if xctype in (None, 'LDA', 'HF'):
        veff = _get_j_pass2(mydf, wv_freq, kpts_band, verbose=log)
    elif xctype == 'GGA':
        if GGA_METHOD.upper() == 'FFT':
            veff = _get_j_pass2(mydf, wv_freq, kpts_band, verbose=log)
        else:
            veff = _get_gga_pass2(mydf, wv_freq, kpts_band, hermi=hermi, verbose=log)
    else:
        raise NotImplementedError

    wv_freq = None
    veff = _format_jks(veff, dm_kpts, input_band, kpts)

    if return_j:
        vj = _get_j_pass2(mydf, vG, kpts_band, verbose=log)
        vj = _format_jks(veff, dm_kpts, input_band, kpts)
    else:
        vj = None
    vG = None

    veff = lib.tag_array(veff, ecoul=ecoul, exc=excsum, vj=vj, vk=None)
    return nelec, excsum, veff

def get_veff_ip1(mydf, dm_kpts, xc_code=None, kpts=np.zeros((1,3)), kpts_band=None, spin=0):
    if isinstance(kpts, KPoints):
        raise NotImplementedError

    cell = mydf.cell
    dm_kpts = np.asarray(dm_kpts, order='C')
    dms = _format_dms(dm_kpts, kpts)
    nset, nkpts, nao = dms.shape[:3]
    kpts_band = _format_kpts_band(kpts_band, kpts)
    if spin == 1:
        nset //= 2

    mesh = mydf.mesh
    ngrids = np.prod(mesh)
    ni = mydf
    xctype = ni._xc_type(xc_code)
    if xctype in (None, 'LDA', 'HF'):
        deriv = 0
    elif xctype == 'GGA':
        deriv = 1
    else:
        raise NotImplementedError
    rhoG = _eval_rhoG(mydf, dm_kpts, hermi=1, kpts=kpts_band, deriv=deriv)
    if spin == 1:
        rhoG = rhoG.reshape(nset,2,-1,ngrids)
    # cache rhoG for core density gradients
    mydf.rhoG = rhoG

    coulG = tools.get_coulG(cell, mesh=mesh)
    vG = np.empty((nset,ngrids), dtype=np.result_type(rhoG, coulG))
    for i in range(nset):
        if spin == 0:
            vG[i] = np.multiply(rhoG[i,0], coulG, out=vG[i])
        elif spin == 1:
            tmp = rhoG[i,0,0] + rhoG[i,1,0]
            vG[i] = np.multiply(tmp, coulG, out=vG[i])

    if mydf.vpplocG_part1 is not None:
        for i in range(nset):
            vG[i] += mydf.vpplocG_part1

    weight = cell.vol / ngrids

    # *(1./weight) because rhoR is scaled by weight in _eval_rhoG.  When
    # computing rhoR with IFFT, the weight factor is not needed.
    rhoR = tools.ifft(rhoG.reshape(-1,ngrids), mesh).real * (1./weight)
    if spin == 0:
        rhoR = rhoR.reshape(nset,-1,ngrids)
    elif spin == 1:
        rhoR = rhoR.reshape(nset,2,-1,ngrids)

    wv_freq = []
    Gv = cell.get_Gv(mesh)
    for i in range(nset):
        vxc = ni.eval_xc_eff(xc_code, rhoR[i], deriv=1, xctype=xctype, spin=spin)[1]
        wv = weight * vxc
        wv = tools.fft(wv.reshape(-1,ngrids), mesh)
        if spin == 0:
            wv = wv.reshape(-1,ngrids)
            if xctype == 'GGA' and GGA_METHOD.upper() == 'FFT':
                wv[0] -= np.einsum('xg,gx->g', wv[1:4], 1j*Gv)
                wv = wv[:1]
        elif spin == 1:
            wv = wv.reshape(2,-1,ngrids)
            if xctype == 'GGA' and GGA_METHOD.upper() == 'FFT':
                wv[:,0] -= np.einsum('nxg,gx->ng', wv[:,1:4], 1j*Gv)
                wv = wv[:,:1]
        wv_freq.append(wv)

    rhoR = rhoG = None
    if spin == 0:
        if len(wv_freq) == 1:
            wv_freq = wv_freq[0].reshape(nset,-1,*mesh)
        else:
            wv_freq = np.asarray(wv_freq).reshape(nset,-1,*mesh)
    elif spin == 1:
        if len(wv_freq) == 1:
            wv_freq = wv_freq[0].reshape(nset,2,-1,*mesh)
        else:
            wv_freq = np.asarray(wv_freq).reshape(nset,2,-1,*mesh)

    for i in range(nset):
        if spin == 0:
            wv_freq[i,0] += vG[i].reshape(*mesh)
        elif spin == 1:
            for s in range(2):
                wv_freq[i,s,0] += vG[i].reshape(*mesh)

    if xctype in (None, 'LDA', 'HF'):
        vj_kpts = _get_j_pass2_ip1(mydf, wv_freq, kpts_band, hermi=0, deriv=1)
    elif xctype == 'GGA':
        if GGA_METHOD.upper() == 'FFT':
            vj_kpts = _get_j_pass2_ip1(mydf, wv_freq, kpts_band, hermi=0, deriv=1)
        else:
            vj_kpts = _get_gga_pass2_ip1(mydf, wv_freq, kpts_band, hermi=0, deriv=1)
    else:
        raise NotImplementedError

    comp = 3
    nao = cell.nao
    if spin == 0:
        vj_kpts = vj_kpts.reshape(nset,nkpts,comp,nao,nao)
    elif spin == 1:
        vj_kpts = vj_kpts.reshape(nset,2,nkpts,comp,nao,nao)
    vj_kpts = np.moveaxis(vj_kpts, -3, -4)

    if nkpts == 1:
        vj_kpts = vj_kpts[...,0,:,:]
    if nset == 1:
        vj_kpts = vj_kpts[0]
    return vj_kpts

def get_nuc(mydf, kpts=None, deriv=0):
    if kpts is None:
        kpts = np.zeros((1, 3))
    elif isinstance(kpts, KPoints):
        kpts = kpts.kpts_ibz
    kpts, is_single_kpt = fft._check_kpts(mydf, kpts)
    cell = mydf.cell
    mesh = mydf.mesh
    charge = -cell.atom_charges()

    Gv, Gvbase, _ = cell.get_Gv_weights(mesh)
    basex, basey, basez = Gvbase
    b = cell.reciprocal_vectors()
    rb = np.dot(cell.atom_coords(), b.T)
    SIx = np.exp(-1j*np.einsum('z,g->zg', rb[:,0], basex))
    SIy = np.exp(-1j*np.einsum('z,g->zg', rb[:,1], basey))
    SIz = np.exp(-1j*np.einsum('z,g->zg', rb[:,2], basez))
    rhoG = np.einsum('i,ix,iy,iz->xyz', charge, SIx, SIy, SIz).ravel()

    coulG = tools.get_coulG(cell, mesh=mesh, Gv=Gv)
    vneG = rhoG * coulG
    if deriv == 0:
        vne = _get_j_pass2(mydf, vneG, kpts=kpts, hermi=1)
    elif deriv == 1: # ip1
        vne = _get_j_pass2_ip1(mydf, vneG, kpts=kpts, hermi=0, deriv=deriv)
    else:
        raise NotImplementedError

    if is_single_kpt:
        vne = vne[0]
    return np.asarray(vne)

def get_nuc_ip1(mydf, kpts=None):
    return get_nuc(mydf, kpts=kpts, deriv=1)

def get_nuc_nuc_grad(mydf, dm, kpts=None):
    # < p | d/dR (-Z / |r-R|) | q > D_{pq}
    if kpts is None:
        kpts = np.zeros((1, 3))
    elif isinstance(kpts, KPoints):
        if kpts.kpts.size > 3: # multiple k points
            dm = kpts.transform_dm(dm)
        kpts = kpts.kpts
    kpts, is_single_kpt = fft._check_kpts(mydf, kpts)
    cell = mydf.cell
    mesh = mydf.mesh
    dms = _format_dms(dm, kpts)

    Gv = cell.get_Gv(mesh)
    coulG = tools.get_coulG(cell, mesh=mesh, Gv=Gv)
    coords = cell.atom_coords()
    charge = -cell.atom_charges()
    grad = np.zeros((cell.natm,3))
    # TODO improve performance
    rhoG = _eval_rhoG(mydf, dms, hermi=1, kpts=kpts).ravel()
    for ia in range(cell.natm):
        vG = 1j * np.exp(1j * np.dot(Gv, coords[ia])) * charge[ia] * coulG * Gv.T
        grad[ia] = np.dot(vG, rhoG).real
    grad *= 1 / cell.vol
    return grad

class MultiGridNumInt(MultiGridNumInt_v1):
    '''Base class for multigrid DFT (version 2).

    Attributes:
        task_list : `TaskList` instance
            Task list recording which primitive basis function pairs
            need to be considered.
        vpplocG_part1 : ndarray
            Long-range part of the local pseudopotential represented
            in the reciprocal space. It is cached to reduce cost.
        rhoG : ndarray
            Electronic density represented in the reciprocal space.
            It is cached in nuclear gradient calculations to reduce cost.
    '''
    ntasks = NTASKS
    ke_ratio = KE_RATIO
    rel_cutoff = REL_CUTOFF
    _keys = {'ntasks', 'ke_ratio', 'rel_cutoff',
             'task_list', 'vpplocG_part1', 'rhoG'}

    def __init__(self, cell):
        MultiGridNumInt_v1.__init__(self, cell)
        self.task_list = None
        self.vpplocG_part1 = None
        self.rhoG = None

    def reset(self, cell=None):
        self.vpplocG_part1 = None
        self.rhoG = None
        if self.task_list is not None:
            self.task_list = None
        MultiGridNumInt_v1.reset(self, cell=cell)
        return self

    @property
    def ngrids(self):
        return np.prod(self.mesh)

    @ngrids.setter
    def ngrids(self, x):
        raise AttributeError('The MultiGridNumInt attribute .ngrids is deprecated. '
                             'It is replaced by the attribute .ntasks.')

    def get_veff_ip1(self, dm, xc_code=None, kpts=None, kpts_band=None, spin=0):
        if kpts is None:
            kpts = np.zeros(1,3)
        kpts = kpts.reshape(-1,3)
        if not gamma_point(kpts):
            raise NotImplementedError('MultiGridNumInt2 only supports Gamma-point calculations.')
        vj = get_veff_ip1(self, dm, xc_code=xc_code,
                          kpts=kpts, kpts_band=kpts_band, spin=spin)
        return vj

    def get_pp(self, kpts=None, return_full=False):
        '''Return the GTH pseudopotential (PP) matrix in AO basis,
        with contribution from G=0 removed.

        By default, the returned PP includes
        the short-range part of the local potential and the non-local potential.
        The long-range part of the local potential is cached as `vpplocG_part1`,
        which is the reciprocal space representation, to be added to the electron
        density Coulomb potential for computing the Coulomb matrix.
        In order to get the full PP matrix, set return_full to True.

        Kwargs:
            return_full: bool
                If True, the returned PP also contains the long-range part.
                Default is False.
        '''
        self.vpplocG_part1 = _get_vpplocG_part1(self, with_rho_core=True)

        vpp = _get_pp_without_erf(self, kpts)
        if return_full:
            if kpts is None:
                kpts = np.zeros((1,3))
            kpts, is_single_kpt = fft._check_kpts(self, kpts)
            vpp1 = _get_j_pass2(self, self.vpplocG_part1, kpts)
            if is_single_kpt:
                vpp1 = vpp1[0]
            vpp += vpp1

        return vpp

    get_nuc = get_nuc
    get_nuc_ip1 = get_nuc_ip1
    get_nuc_nuc_grad = get_nuc_nuc_grad
    vpploc_part1_nuc_grad = vpploc_part1_nuc_grad
    get_vpploc_part1_ip1 = get_vpploc_part1_ip1

    def nr_rks(self, cell, grids, xc_code, dm_kpts, relativity=0, hermi=1,
               kpts=None, kpts_band=None, max_memory=2000, verbose=None):
        if kpts is not None and not gamma_point(kpts):
            raise NotImplementedError('MultiGridNumInt2 only supports Gamma-point calculations.')
        return_j = False
        return nr_rks(self, xc_code, dm_kpts, hermi, kpts, kpts_band, self.xc_with_j,
                      return_j, verbose)

    def nr_uks(self, cell, grids, xc_code, dm_kpts, relativity=0, hermi=1,
               kpts=None, kpts_band=None, max_memory=2000, verbose=None):
        if kpts is not None and not gamma_point(kpts):
            raise NotImplementedError('MultiGridNumInt2 only supports Gamma-point calculations.')
        return_j = False
        return nr_uks(self, xc_code, dm_kpts, hermi, kpts, kpts_band, self.xc_with_j,
                      return_j, verbose)

    nr_rks_fxc = NotImplemented
    nr_uks_fxc = NotImplemented
    nr_rks_fxc_st = NotImplemented
    cache_xc_kernel = NotImplemented
    cache_xc_kernel1 = NotImplemented
    get_rho = NotImplemented



#!/usr/bin/env python
# Copyright 2014-2023 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Timothy Berkelbach <tim.berkelbach@gmail.com>
#

'''
parse CP2K format
'''

import re
from pyscf.lib.exceptions import BasisNotFoundError
from pyscf.gto.basis import parse_nwchem
from pyscf.gto.basis.parse_nwchem import _search_basis_block
from pyscf import __config__

DISABLE_EVAL = getattr(__config__, 'DISABLE_EVAL', False)

MAXL = 8

def parse(string, symb=None, optimize=False):
    '''Parse the basis text which is in CP2K format, return an internal
    basis format which can be assigned to :attr:`Mole.basis`
    Lines started with # are ignored.

    Examples:

    >>> cell = gto.Cell()
    >>> cell.basis = {'C': pyscf.gto.basis.parse_cp2k.parse("""
    ... C DZVP-GTH
    ...   2
    ...   2  0  1  4  2  2
    ...         4.3362376436   0.1490797872   0.0000000000  -0.0878123619   0.0000000000
    ...         1.2881838513  -0.0292640031   0.0000000000  -0.2775560300   0.0000000000
    ...         0.4037767149  -0.6882040510   0.0000000000  -0.4712295093   0.0000000000
    ...         0.1187877657  -0.3964426906   1.0000000000  -0.4058039291   1.0000000000
    ...   3  2  2  1  1
    ...         0.5500000000   1.0000000000
    ... #
    ... """)}
    '''
    if symb is not None:
        raw_data = list(filter(None, re.split(BASIS_SET_DELIMITER, string)))
        line_data = _search_basis_block(raw_data, symb)
        if not line_data:
            raise BasisNotFoundError(f'Basis not found for {symb}')
    else:
        line_data = string.splitlines()

    bastxt = []
    for dat in line_data:
        x = dat.split('#')[0].strip()
        if (x and not x.startswith('END') and not x.startswith('BASIS')):
            bastxt.append(x)
    return _parse(bastxt, optimize)

def load(basisfile, symb, optimize=False):
    return _parse(search_seg(basisfile, symb), optimize)

def _parse(blines, optimize=False):
    blines_iter = iter(blines)
    try:
        header_ln = next(blines_iter)  # noqa: F841
        nsets = int(next(blines_iter))
    except Exception:
        raise BasisNotFoundError('Not basis data')

    basis = []
    try:
        for n in range(nsets):
            comp = [int(p) for p in next(blines_iter).split()]
            lmin, lmax, nexps, ncontractions = comp[1], comp[2], comp[3], comp[4:]
            basis_n = [[l] for l in range(lmin,lmax+1)]
            for nexp in range(nexps):
                line = next(blines_iter)
                dat = line.split()
                try:
                    bfun = [float(x) for x in dat]
                except ValueError:
                    if DISABLE_EVAL:
                        raise ValueError('Failed to parse %s' % line)
                    else:
                        bfun = eval(','.join(dat))

                if len(bfun) != sum(ncontractions) + 1:
                    raise ValueError('Basis data incomplete')

                bfun_iter = iter(bfun)
                exp = next(bfun_iter)
                for i,l in enumerate(range(lmin,lmax+1)):
                    cl = [exp]
                    for c in range(ncontractions[i]):
                        cl.append(next(bfun_iter))
                    basis_n[i].append(cl)
            basis.extend(basis_n)
    except StopIteration:
        raise ValueError('Basis data incomplete')

    basis_sorted = []
    for l in range(MAXL):
        basis_sorted.extend([b for b in basis if b[0] == l])
    if not basis_sorted:
        raise BasisNotFoundError('Basis data not found')

    if optimize:
        basis_sorted = parse_nwchem.optimize_contraction(basis_sorted)

    basis_sorted = parse_nwchem.remove_zero(basis_sorted)
    return basis_sorted

BASIS_SET_DELIMITER = re.compile('# *BASIS SET.*\n')
def search_seg(basisfile, symb):
    with open(basisfile, 'r') as fin:
        fdata = re.split(BASIS_SET_DELIMITER, fin.read())
    line_data = _search_basis_block(fdata[1:], symb)
    if not line_data:
        raise BasisNotFoundError(f'Basis for {symb} not found in {basisfile}')
    return [x for x in line_data if x and 'END' not in x]



# Copyright 2014-2022 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Samragni Banerjee <samragnibanerjee4@gmail.com>
#         Alexander Sokolov <alexander.y.sokolov@gmail.com>
#

import numpy as np
import pyscf.ao2mo as ao2mo
from pyscf import lib
from pyscf.lib import logger
from pyscf.pbc.df import df
from pyscf.pbc.mp.kmp2 import (get_frozen_mask, get_nocc, get_nmo,
                               padded_mo_coeff, padding_k_idx)
import time
import h5py
import tempfile

### Incore integral transformation for integrals in Chemists' notation###
def transform_integrals_incore(myadc):

    log = logger.Logger(myadc.stdout, myadc.verbose)
    kpts = myadc.kpts
    nkpts = myadc.nkpts
    nocc = myadc.nocc
    nmo = myadc.nmo
    nvir = nmo - nocc
    dtype = myadc.mo_coeff[0].dtype

    mo_coeff = myadc.mo_coeff = padded_mo_coeff(myadc, myadc.mo_coeff)

    fao2mo = myadc._scf.with_df.ao2mo

    kconserv = myadc.khelper.kconserv
    khelper = myadc.khelper

    orbv = np.asarray(mo_coeff[:,:,nocc:], order='C')

    fao2mo = myadc._scf.with_df.ao2mo
    eris = lambda:None

    log.info('using incore ERI storage')
    eris.oooo = np.empty((nkpts,nkpts,nkpts,nocc,nocc,nocc,nocc), dtype=dtype)
    eris.oovv = np.empty((nkpts,nkpts,nkpts,nocc,nocc,nvir,nvir), dtype=dtype)
    eris.ovoo = np.empty((nkpts,nkpts,nkpts,nocc,nvir,nocc,nocc), dtype=dtype)
    eris.ovov = np.empty((nkpts,nkpts,nkpts,nocc,nvir,nocc,nvir), dtype=dtype)
    eris.ovvv = np.empty((nkpts,nkpts,nkpts,nocc,nvir,nvir,nvir), dtype=dtype)
    eris.ovvo = np.empty((nkpts,nkpts,nkpts,nocc,nvir,nvir,nocc), dtype=dtype)

    for (ikp,ikq,ikr) in khelper.symm_map.keys():
        iks = kconserv[ikp,ikq,ikr]
        eri_kpt = fao2mo((mo_coeff[ikp],mo_coeff[ikq],mo_coeff[ikr],mo_coeff[iks]),
                         (kpts[ikp],kpts[ikq],kpts[ikr],kpts[iks]), compact=False)
        if dtype == np.float64:
            eri_kpt = eri_kpt.real
        eri_kpt = eri_kpt.reshape(nmo, nmo, nmo, nmo)
        for (kp, kq, kr) in khelper.symm_map[(ikp, ikq, ikr)]:
            eri_kpt_symm = khelper.transform_symm(eri_kpt, kp, kq, kr)
            eris.oooo[kp,kq,kr] = eri_kpt_symm[:nocc,:nocc,:nocc,:nocc]/nkpts
            eris.oovv[kp,kq,kr] = eri_kpt_symm[:nocc,:nocc, nocc:,nocc:]/nkpts
            eris.ovoo[kp,kq,kr] = eri_kpt_symm[:nocc,nocc:,:nocc,:nocc]/nkpts
            eris.ovov[kp,kq,kr] = eri_kpt_symm[:nocc,nocc:,:nocc,nocc:]/nkpts
            eris.ovvv[kp,kq,kr] = eri_kpt_symm[:nocc,nocc:,nocc:,nocc:]/nkpts
            eris.ovvo[kp,kq,kr] = eri_kpt_symm[:nocc,nocc:,nocc:,:nocc]/nkpts

    if (myadc.method == "adc(2)-x" and myadc.higher_excitations is True) or (myadc.method == "adc(3)"):
        eris.vvvv = myadc._scf.with_df.ao2mo_7d(orbv, factor=1./nkpts).transpose(0,2,1,3,5,4,6)

    return eris


def transform_integrals_outcore(myadc):

    from pyscf.pbc import tools
    from pyscf.pbc.cc.ccsd import _adjust_occ

    log = logger.Logger(myadc.stdout, myadc.verbose)
    kpts = myadc.kpts
    nkpts = myadc.nkpts
    nocc = myadc.nocc
    nmo = myadc.nmo
    nvir = nmo - nocc

    dtype = myadc.mo_coeff[0].dtype

    mo_coeff = myadc.mo_coeff = padded_mo_coeff(myadc, myadc.mo_coeff)

    fao2mo = myadc._scf.with_df.ao2mo

    kconserv = myadc.khelper.kconserv
    khelper = myadc.khelper

    eris = lambda:None
    eris.feri = feri = lib.H5TmpFile()

    # The momentum conservation array
    kconserv = myadc.khelper.kconserv

    eris.feri = feri = lib.H5TmpFile()
    eris.oooo = feri.create_dataset('oooo', (nkpts,nkpts,nkpts,nocc,nocc,nocc,nocc), dtype=dtype)
    eris.oovv = feri.create_dataset('oovv', (nkpts,nkpts,nkpts,nocc,nocc,nvir,nvir), dtype=dtype)
    eris.ovoo = feri.create_dataset('ovoo', (nkpts,nkpts,nkpts,nocc,nvir,nocc,nocc), dtype=dtype)
    eris.ovov = feri.create_dataset('ovov', (nkpts,nkpts,nkpts,nocc,nvir,nocc,nvir), dtype=dtype)
    eris.ovvv = feri.create_dataset('ovvv', (nkpts,nkpts,nkpts,nocc,nvir,nvir,nvir), dtype=dtype)
    eris.ovvo = feri.create_dataset('ovvo', (nkpts,nkpts,nkpts,nocc,nvir,nvir,nocc), dtype=dtype)
    eris.vvvv = feri.create_dataset('vvvv', (nkpts,nkpts,nkpts,nvir,nvir,nvir,nvir), dtype=dtype)

    cput1 = time.process_time(), time.time()
    for kp in range(nkpts):
        for kq in range(nkpts):
            for kr in range(nkpts):
                ks = kconserv[kp, kq, kr]
                orbo_p = mo_coeff[kp][:, :nocc]
                orbo_q = mo_coeff[kq][:, :nocc]
                buf_kpt = fao2mo((orbo_p, orbo_q, mo_coeff[kr], mo_coeff[ks]),
                                 (kpts[kp], kpts[kq], kpts[kr], kpts[ks]), compact=False)
                if mo_coeff[0].dtype == np.float64:
                    buf_kpt = buf_kpt.real
                buf_kpt = buf_kpt.reshape(nocc, nocc, nmo, nmo)
                dtype = buf_kpt.dtype
                eris.oooo[kp, kq, kr, :, :, :, :] = buf_kpt[:, :, :nocc, :nocc] / nkpts
                eris.oovv[kp, kq, kr, :, :, :, :] = buf_kpt[:, :, nocc:, nocc:] / nkpts
    cput1 = log.timer_debug1('transforming oopq', *cput1)

    # <ia|pq> = (ip|aq)
    cput1 = time.process_time(), time.time()
    for kp in range(nkpts):
        for kq in range(nkpts):
            for kr in range(nkpts):
                ks = kconserv[kp, kq, kr]
                orbo_p = mo_coeff[kp][:, :nocc]
                orbv_q = mo_coeff[kq][:, nocc:]
                buf_kpt = fao2mo((orbo_p, orbv_q, mo_coeff[kr], mo_coeff[ks]),
                                 (kpts[kp], kpts[kq], kpts[kr], kpts[ks]), compact=False)
                if mo_coeff[0].dtype == np.float64:
                    buf_kpt = buf_kpt.real
                buf_kpt = buf_kpt.reshape(nocc,nvir,nmo, nmo)
                eris.ovoo[kp, kq, kr, :, :, :, :] = buf_kpt[:, :, :nocc, :nocc] / nkpts
                eris.ovov[kp, kq, kr, :, :, :, :] = buf_kpt[:, :, :nocc, nocc:] / nkpts
                eris.ovvo[kp, kq, kr, :, :, :, :] = buf_kpt[:, :, nocc:, :nocc] / nkpts
                eris.ovvv[kp, kq, kr, :, :, :, :] = buf_kpt[:, :, nocc:, nocc:] / nkpts
            cput1 = log.timer_debug1('transforming ovpq', *cput1)

    if (myadc.method == "adc(2)-x" and myadc.higher_excitations is True) or (myadc.method == "adc(3)"):
        mem_now = lib.current_memory()[0]
        if nvir ** 4 * 16 / 1e6 + mem_now < myadc.max_memory:
            for (ikp, ikq, ikr) in khelper.symm_map.keys():
                iks = kconserv[ikp, ikq, ikr]
                orbv_p = mo_coeff[ikp][:, nocc:]
                orbv_q = mo_coeff[ikq][:, nocc:]
                orbv_r = mo_coeff[ikr][:, nocc:]
                orbv_s = mo_coeff[iks][:, nocc:]
                # unit cell is small enough to handle vvvv in-core
                buf_kpt = fao2mo((orbv_p,orbv_q,orbv_r,orbv_s),
                                 kpts[[ikp,ikq,ikr,iks]], compact=False)
                if dtype == np.float64:
                    buf_kpt = buf_kpt.real
                buf_kpt = buf_kpt.reshape((nvir, nvir, nvir, nvir))
                for (kp, kq, kr) in khelper.symm_map[(ikp, ikq, ikr)]:
                    buf_kpt_symm = khelper.transform_symm(buf_kpt, kp, kq, kr).transpose(0, 2, 1, 3)
                    eris.vvvv[kp, kr, kq] = buf_kpt_symm / nkpts
        else:
            #raise MemoryError('Minimal memory requirements %s MB'
            #                  % (mem_now + nvir ** 4 / 1e6 * 16 * 2))
            for (ikp, ikq, ikr) in khelper.symm_map.keys():
                for a in range(nvir):
                    orbva_p = orbv_p[:, a].reshape(-1, 1)
                    buf_kpt = fao2mo((orbva_p, orbv_q, orbv_r, orbv_s),
                                     (kpts[ikp], kpts[ikq], kpts[ikr], kpts[iks]), compact=False)
                    if mo_coeff[0].dtype == np.float64:
                        buf_kpt = buf_kpt.real
                    buf_kpt = buf_kpt.reshape((1, nvir, nvir, nvir)).transpose(0, 2, 1, 3)

                    eris.vvvv[ikp, ikr, ikq, a, :, :, :] = buf_kpt[0, :, :, :] / nkpts
                    # Store symmetric permutations
                    eris.vvvv[ikr, ikp, iks, :, a, :, :] = buf_kpt.transpose(1, 0, 3, 2)[
                        :, 0, :, :] / nkpts
                    eris.vvvv[ikq, iks, ikp, :, :, a, :] = buf_kpt.transpose(2, 3, 0, 1).conj()[
                        :, :, 0, :] / nkpts
                    eris.vvvv[iks, ikq, ikr, :, :, :, a] = buf_kpt.transpose(3, 2, 1, 0).conj()[
                        :, :, :, 0] / nkpts
            cput1 = log.timer_debug1('transforming vvvv', *cput1)

    return eris


def transform_integrals_df(myadc):
    from pyscf.ao2mo import _ao2mo
    cell = myadc.cell
    kpts = myadc.kpts
    nkpts = myadc.nkpts
    nocc = myadc.nocc
    nmo = myadc.nmo
    nvir = nmo - nocc
    nao = cell.nao_nr()

    if myadc._scf.with_df._cderi is None:
        myadc._scf.with_df.build()
    dtype = myadc.mo_coeff[0].dtype

    mo_coeff = myadc.mo_coeff = padded_mo_coeff(myadc, myadc.mo_coeff)

    kconserv = myadc.khelper.kconserv

    # The momentum conservation array
    kconserv = myadc.khelper.kconserv

    with_df = myadc.with_df
    naux = with_df.get_naoaux()
    eris = lambda:None

    eris.dtype = dtype = np.result_type(dtype)
    eris.Lpq_mo = Lpq_mo = np.empty((nkpts, nkpts), dtype=object)
    Loo = np.empty((nkpts,nkpts,naux,nocc,nocc),dtype=dtype)
    Lvo = np.empty((nkpts,nkpts,naux,nvir,nocc),dtype=dtype)
    eris.Lvv = np.empty((nkpts,nkpts,naux,nvir,nvir),dtype=dtype)
    eris.Lov = np.empty((nkpts,nkpts,naux,nocc,nvir),dtype=dtype)

    eris.vvvv = None
    eris.ovvv = None

    with df._load3c(myadc._scf.with_df._cderi, 'j3c') as fload:
        tao = []
        ao_loc = None
        for ki, kpti in enumerate(kpts):
            for kj, kptj in enumerate(kpts):
                Lpq_ao = np.asarray(fload(kpti, kptj))

                mo = np.hstack((mo_coeff[ki], mo_coeff[kj]))
                mo = np.asarray(mo, dtype=dtype, order='F')
                if dtype == np.double:
                    out = _ao2mo.nr_e2(Lpq_ao, mo, (0, nmo, nmo, nmo+nmo), aosym='s2')
                else:
                    #Note: Lpq.shape[0] != naux if linear dependency is found in auxbasis
                    if Lpq_ao[0].size != nao**2:  # aosym = 's2'
                        Lpq_ao = lib.unpack_tril(Lpq_ao).astype(np.complex128)
                    out = _ao2mo.r_e2(Lpq_ao, mo, (0, nmo, nmo, nmo+nmo), tao, ao_loc)
                Lpq_mo[ki, kj] = out.reshape(-1, nmo, nmo)

                Loo[ki,kj] = Lpq_mo[ki,kj][:,:nocc,:nocc]
                eris.Lov[ki,kj] = Lpq_mo[ki,kj][:,:nocc,nocc:]
                Lvo[ki,kj] = Lpq_mo[ki,kj][:,nocc:,:nocc]
                eris.Lvv[ki,kj] = Lpq_mo[ki,kj][:,nocc:,nocc:]

    eris.feri = feri = lib.H5TmpFile()

    eris.oooo = feri.create_dataset('oooo', (nkpts,nkpts,nkpts,nocc,nocc,nocc,nocc), dtype=dtype)
    eris.oovv = feri.create_dataset('oovv', (nkpts,nkpts,nkpts,nocc,nocc,nvir,nvir), dtype=dtype)
    eris.ovoo = feri.create_dataset('ovoo', (nkpts,nkpts,nkpts,nocc,nvir,nocc,nocc), dtype=dtype)
    eris.ovov = feri.create_dataset('ovov', (nkpts,nkpts,nkpts,nocc,nvir,nocc,nvir), dtype=dtype)
    eris.ovvo = feri.create_dataset('ovvo', (nkpts,nkpts,nkpts,nocc,nvir,nvir,nocc), dtype=dtype)
    #eris.ovvv = feri.create_dataset('ovvv', (nkpts,nkpts,nkpts,nocc,nvir,nvir,nvir), dtype=dtype)

    for kp in range(nkpts):
        for kq in range(nkpts):
            for kr in range(nkpts):
                ks = kconserv[kp,kq,kr]
                eris.oooo[kp,kq,kr] = lib.einsum('Lpq,Lrs->pqrs', Loo[kp,kq], Loo[kr,ks])/nkpts
                eris.oovv[kp,kq,kr] = lib.einsum('Lpq,Lrs->pqrs', Loo[kp,kq], eris.Lvv[kr,ks])/nkpts
                eris.ovoo[kp,kq,kr] = lib.einsum('Lpq,Lrs->pqrs', eris.Lov[kp,kq], Loo[kr,ks])/nkpts
                eris.ovov[kp,kq,kr] = lib.einsum(
                    'Lpq,Lrs->pqrs', eris.Lov[kp,kq], eris.Lov[kr,ks])/nkpts
                eris.ovvo[kp,kq,kr] = lib.einsum('Lpq,Lrs->pqrs', eris.Lov[kp,kq], Lvo[kr,ks])/nkpts
                #eris.ovvv[kp,kq,kr] = lib.einsum('Lpq,Lrs->pqrs', eris.Lov[kp,kq], Lvv[kr,ks])/nkpts

    return eris

def calculate_chunk_size(myadc):

    avail_mem = (myadc.max_memory - lib.current_memory()[0]) * 0.5
    nocc = [np.count_nonzero(myadc.mo_occ[ikpt]) for ikpt in range(myadc.nkpts)]
    nocc = np.amax(nocc)
    nmo = [len(myadc.mo_occ[ikpt]) for ikpt in range(myadc.nkpts)]
    nmo = np.max(nocc) + np.max(np.array(nmo) - np.array(nocc))
    nvir = nmo - nocc
    vvv_mem = (nvir**3) * 8/1e6

    chnk_size =  int(avail_mem/vvv_mem)

    if chnk_size <= 0 :
        chnk_size = 1

    return chnk_size



#!/usr/bin/env python
# Copyright 2022-2023 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Authors: Xing Zhang <zhangxing.nju@gmail.com>
#

import warnings
from functools import reduce
import numpy as np
from numpy.lib.mixins import NDArrayOperatorsMixin
from pyscf import lib


def empty(shape, dtype=float, order='C', metadata=None):
    if metadata is None:
        return np.empty(shape, dtype, order)
    else:
        return KsymmArray(shape, dtype, order, metadata)

def empty_like(a, *args, **kwargs):
    if isinstance(a, KsymmArray):
        return KsymmArray(a.subarray_shape,
                          a.dtype,
                          a.subarray_order,
                          a.metadata)
    else:
        return np.empty_like(a)


class KsymmArray(NDArrayOperatorsMixin):
    def __init__(self, subarray_shape, dtype=float, subarray_order='C', metadata={},
                 init_with_zeros=False):
        self.metadata = metadata
        self._subarray_shape = list(subarray_shape)
        self._subarray_ndim = len(subarray_shape)
        self._subarray_order = subarray_order
        self._dtype = np.dtype(dtype)
        incore = metadata.get('incore', True)
        self._datafile = None
        self.data = self._init(subarray_order, incore, init_with_zeros)

    def _init(self, order, incore=True, init_with_zeros=False):
        if self.subarray_ndim == 2:
            kpts = self.metadata['kpts']
            n_subarray = kpts.nkpts_ibz
        elif self.subarray_ndim == 4:
            kqrts = self.metadata['kqrts']
            n_subarray = len(kqrts.kqrts_ibz)
        else:
            raise NotImplementedError

        data = None
        shape = [n_subarray,] + self.subarray_shape
        if incore:
            if init_with_zeros:
                fn_init = np.zeros
            else:
                fn_init = np.empty
            if order == 'C':
                data = fn_init(shape, self.dtype, order)
            else:
                data = []
                for i in range(n_subarray):
                    data.append(fn_init(self.subarray_shape, self.dtype, order))
                data = np.asarray(data, order='K')
        else:
            self._datafile = lib.H5TmpFile()
            data = self._datafile.create_dataset('data', shape, self.dtype.char)
        return data

    @property
    def shape(self):
        nkpts = self.metadata['kpts'].nkpts
        nk = [nkpts,] * (self.subarray_ndim-1)
        return tuple(nk + list(self.subarray_shape))

    @property
    def ndim(self):
        return self.subarray_ndim-1 + self.subarray_ndim

    @property
    def subarray_ndim(self):
        return self._subarray_ndim

    @property
    def subarray_shape(self):
        return self._subarray_shape

    @property
    def subarray_order(self):
        return self._subarray_order

    @property
    def dtype(self):
        return self._dtype

    def __getitem__(self, key):
        if self.subarray_ndim == 2:
            return self._getitem_2d(key)
        elif self.subarray_ndim == 4:
            return self._getitem_4d(key)
        else:
            raise NotImplementedError

    def __setitem__(self, key, value):
        if self.subarray_ndim == 2:
            return self._setitem_2d(key, value)
        elif self.subarray_ndim == 4:
            return self._setitem_4d(key, value)
        else:
            raise NotImplementedError

    def _getitem_2d(self, key):
        kpts = self.metadata['kpts']
        rmat = self.metadata['rmat']
        label = self.metadata['label']
        trans = self.metadata['trans']
        if isinstance(key, (int, np.integer)):
            return transform_2d(self.data, kpts, key, rmat, label, trans)
        elif isinstance(key, (slice, np.ndarray)):
            data = []
            for ki in np.arange(kpts.nkpts)[key]:
                data.append(transform_2d(self.data, kpts, ki, rmat, label, trans))
            return np.asarray(data)
        else:
            raise NotImplementedError

    def _getitem_4d(self, key):
        kpts = self.metadata['kpts']
        kqrts = self.metadata['kqrts']
        rmat = self.metadata['rmat']
        label = self.metadata['label']
        trans = self.metadata['trans']

        shape = [kpts.nkpts,] * 3
        coords = index_to_coords(key, shape)
        if coords.ndim == 1:
            klc = coords
            return transform_4d(self.data, kpts, kqrts, klc, rmat, label, trans)
        else:
            data = []
            for klc in coords:
                data.append(transform_4d(self.data, kpts, kqrts, klc, rmat, label, trans))
            return np.asarray(data)

    def _setitem_2d(self, key, value):
        kpts = self.metadata['kpts']
        #TODO allow broadcasting
        value = value.reshape(-1, *self.subarray_shape)
        if isinstance(key, (int, np.integer)):
            set_2d(self.data, value, kpts, key)
        elif isinstance(key, (slice, np.ndarray)):
            ki = np.arange(kpts.nkpts)[key]
            set_2d(self.data, value, kpts, ki)
        else:
            raise NotImplementedError

    def _setitem_4d(self, key, value):
        kpts = self.metadata['kpts']
        kqrts = self.metadata['kqrts']
        #TODO allow broadcasting
        value = value.reshape(-1, *self.subarray_shape)

        shape = [kpts.nkpts,] * 3
        coords = index_to_coords(key, shape)
        set_4d(self.data, value, kpts, kqrts, coords)

    def todense(self):
        #TODO allow to return a hdf5 dataset
        return self[:].reshape(self.shape)

    @staticmethod
    def fromdense(arr, shape, dtype=None, order=None, metadata=None):
        if dtype is None:
            dtype = arr.dtype
        order = _guess_input_order(arr, order)
        if metadata is None:
            raise RuntimeError('metadata not initialized')

        out = KsymmArray(shape, dtype, order, metadata)
        arr = arr.reshape(out.shape)
        if out.subarray_ndim == 2:
            kpts = out.metadata['kpts']
            for ki in kpts.ibz2bz:
                ki_ibz = kpts.bz2ibz[ki]
                out[ki_ibz] = arr[ki]
        elif out.subarray_ndim == 4:
            kqrts = out.metadata['kqrts']
            for m, kq in enumerate(kqrts.kqrts_ibz):
                ki, kj, ka, kb = kq
                out[m] = arr[ki, kj, ka]
        else:
            raise NotImplementedError
        return out

    @staticmethod
    def fromraw(arr, shape, dtype=None, order=None, metadata=None):
        if dtype is None:
            dtype = arr.dtype
        order = _guess_input_order(arr, order)
        if metadata is None:
            raise RuntimeError('metadata not initialized')

        out = KsymmArray(shape, dtype, order, metadata)
        arr = arr.reshape(-1, *out.subarray_shape)
        for i, a in enumerate(arr):
            out.data[i] = np.asarray(a, dtype=dtype, order=order)
        return out

    @staticmethod
    def zeros(shape, dtype=float, order='C', metadata=None):
        out = KsymmArray(shape, dtype, order, metadata, init_with_zeros=True)
        return out


def _guess_input_order(arr, order=None):
    if order is None:
        order = 'C'
        if isinstance(arr, np.ndarray):
            if arr.flags.c_contiguous:
                order = 'C'
            elif arr.flags.f_contiguous:
                order = 'F'
    elif order not in ('C', 'F'):
        order = 'C'
    return order

def set_2d(arr, value, kpts, ki):
    if isinstance(ki, (int, np.integer)):
        ki = np.asarray([ki,])

    mask = np.isin(ki, kpts.ibz2bz)
    if not mask.all():
        warnings.warn(f'Indices {ki[~mask]} are not in the irreducible wedge. '
                       'The corresponding data will be discarded.')

    ki_ibz = kpts.bz2ibz[ki[mask]]
    arr[ki_ibz] = value[mask]

def set_4d(arr, value, kpts, kqrts, klc):
    klc = klc.reshape(-1, 3)
    kk_bz = [kpts.ktuple_to_index(s) for s in klc]
    kk_bz = np.asarray(kk_bz)

    mask = np.isin(kk_bz, kqrts.ibz2bz)
    if not mask.all():
        kk_tmp = [kpts.index_to_ktuple(k, 3) for k in kk_bz[~mask]]
        warnings.warn(f'Indices {kk_tmp} are not in the irreducible wedge. '
                       'The corresponding data will be discarded.')

    kk_ibz = kqrts.bz2ibz[kk_bz[mask]]
    arr[kk_ibz] = value[mask]

def transform_2d(arr, kpts, ki, rmat, label, trans):
    ki_ibz = kpts.bz2ibz[ki]
    ki_ibz_bz = kpts.ibz2bz[ki_ibz]
    if ki == ki_ibz_bz:
        return arr[ki_ibz]

    pi, pj = label
    rmat_i = getattr(rmat, pi*2)
    rmat_j = getattr(rmat, pj*2)

    iop = kpts.stars_ops_bz[ki]
    rot_i = rmat_i[ki_ibz_bz][iop]
    rot_j = rmat_j[ki_ibz_bz][iop]
    ti, tj = trans
    if ti == 'c':
        rot_i = rot_i.conj()
    if tj == 'c':
        rot_j = rot_j.conj()

    #:einsum('ia,ij,ab->jb', arr[ki_ibz], rot_i, rot_j)
    out = reduce(np.dot, (rot_i.T, arr[ki_ibz], rot_j))
    return out

def transform_4d(arr, kpts, kqrts, klc, rmat, label, trans):
    kk_bz = kpts.ktuple_to_index(klc)
    kk_ibz = kqrts.bz2ibz[kk_bz]
    i,j,a,b = kqrts.kqrts_ibz[kk_ibz]
    if (i,j,a) == tuple(klc):
        return arr[kk_ibz]

    pi, pj, pa, pb = label
    rmat_i = getattr(rmat, pi*2)
    rmat_j = getattr(rmat, pj*2)
    rmat_a = getattr(rmat, pa*2)
    rmat_b = getattr(rmat, pb*2)

    iop = kqrts.stars_ops_bz[kk_bz]
    rot_i = rmat_i[i][iop]
    rot_j = rmat_j[j][iop]
    rot_a = rmat_a[a][iop]
    rot_b = rmat_b[b][iop]

    ti, tj, ta, tb = trans
    if ti == 'c':
        rot_i = rot_i.conj()
    if tj == 'c':
        rot_j = rot_j.conj()
    if ta == 'c':
        rot_a = rot_a.conj()
    if tb == 'c':
        rot_b = rot_b.conj()

    di, dj, da, db = arr[kk_ibz].shape
    #:einsum('ijab,ik,jl,ac,bd->klcd', arr[kk_ibz],
    #        rot_i, rot_j, rot_a, rot_b)

    #tmp = np.einsum('ik,ijab->jkab', rot_i, arr[kk_ibz])
    tmp = np.dot(rot_i.T, arr[kk_ibz].reshape(di,-1)) #k,jab
    tmp = tmp.reshape(di,dj,-1).transpose(1,0,2) #j,k,ab

    #tmp = np.einsum('jkab,jl->klab', tmp, rot_j)
    tmp = np.dot(rot_j.T, tmp.reshape(dj,-1)) #l,kab
    tmp = tmp.reshape(dj,di,-1).transpose(1,0,2) #k,l,ab

    #tmp = np.einsum('klab,ac->klbc', tmp, rot_a)
    tmp = tmp.reshape(-1,da,db).transpose(0,2,1).reshape(-1,da) #klb,a
    tmp = np.dot(tmp, rot_a) #klb,c

    #out = np.einsum('klbc,bd->klcd', tmp, rot_b)
    tmp = tmp.reshape(-1,db,da).transpose(0,2,1).reshape(-1,db) #klc,b
    out = np.dot(tmp, rot_b).reshape(di,dj,da,db) #k,l,c,d
    return out

def index_to_coords(key, shape):
    if not isinstance(key, tuple):
        key = (key,)

    idxs = []
    for i, k in enumerate(key):
        n = shape[i]
        if isinstance(k, slice):
            idx = slice_to_coords(k, n)
        elif isinstance(k, (int, np.integer)):
            idx = [k,]
        elif isinstance(k, np.ndarray) and k.ndim == 1:
            idx = k
        else:
            raise NotImplementedError
        idxs.append(idx)

    ndim = len(shape)
    if len(idxs) > ndim:
        raise RuntimeError
    elif len(idxs) < ndim:
        for i in range(len(idxs),ndim):
            n = shape[i]
            idxs.append(np.arange(n))

    coords = lib.cartesian_prod(idxs)
    if all(isinstance(k, (int, np.integer)) for k in key) and len(key) == ndim:
        coords = coords[0]
    return coords

def slice_to_coords(k, n):
    start, stop, step = k.start, k.stop, k.step
    if start is None:
        start = 0
    elif start < 0:
        start += n
    if stop is None:
        stop = n
    elif stop < 0:
        stop += n
    if step is None:
        step = 1
    return np.arange(start, stop, step)


zeros = KsymmArray.zeros
fromraw = KsymmArray.fromraw
fromdense = KsymmArray.fromdense



#!/usr/bin/env python
# Copyright 2014-2024 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import numpy as np
from pyscf import mcpdft
from pyscf.grad import mspdft as mspdft_grad
from pyscf import lib
from pyscf.fci import direct_spin1
from pyscf.nac import sacasscf as sacasscf_nacs
from functools import reduce

_unpack_state = mspdft_grad._unpack_state
_nac_csf = sacasscf_nacs._nac_csf

def nac_model (mc_grad, mo_coeff=None, ci=None, si_bra=None, si_ket=None,
               mf_grad=None, atmlst=None):
    '''Compute the "model-state contribution" to the MS-PDFT NAC'''
    mc = mc_grad.base
    mol = mc.mol
    ci_bra = np.tensordot (si_bra, ci, axes=1)
    ci_ket = np.tensordot (si_ket, ci, axes=1)
    ncore, ncas, nelecas = mc.ncore, mc.ncas, mc.nelecas
    castm1 = direct_spin1.trans_rdm1 (ci_bra, ci_ket, ncas, nelecas)
    # if PySCF commentary is to be trusted, trans_rdm1[p,q] is
    # <bra|q'p|ket>. I want <bra|p'q - q'p|ket>.
    castm1 = castm1.conj ().T - castm1
    mo_cas = mo_coeff[:,ncore:][:,:ncas]
    tm1 = reduce (np.dot, (mo_cas, castm1, mo_cas.conj ().T))
    return _nac_csf (mol, mf_grad, tm1, atmlst)

class NonAdiabaticCouplings (mspdft_grad.Gradients):
    '''MS-PDFT non-adiabatic couplings (NACs) between states

    kwargs/attributes:

    state : tuple of length 2
        The NACs returned are <state[1]|d(state[0])/dR>.
        In other words, state = (ket, bra).
    mult_ediff : logical
        If True, returns NACs multiplied by the energy difference.
        Useful near conical intersections to avoid numerical problems.
    use_etfs : logical
        If True, use the ``electron translation factors'' of Fatehi and
        Subotnik [JPCL 3, 2039 (2012)], which guarantee conservation of
        total electron + nuclear momentum when the nuclei are moving
        (i.e., in non-adiabatic molecular dynamics). This corresponds
        to omitting the ``model state contribution''.
    '''

    def __init__(self, mc, state=None, mult_ediff=False, use_etfs=False):
        self.mult_ediff = mult_ediff
        self.use_etfs = use_etfs
        self.state = state
        mspdft_grad.Gradients.__init__(self, mc)

    def get_wfn_response (self, si_bra=None, si_ket=None, state=None, si=None,
                          verbose=None, **kwargs):
        g_all = mspdft_grad.Gradients.get_wfn_response (
            self, si_bra=si_bra, si_ket=si_ket, state=state, si=si,
            verbose=verbose, **kwargs
        )
        g_orb, g_ci, g_is = self.unpack_uniq_var (g_all)
        if state is None: state = self.state
        ket, bra = _unpack_state (state)
        if si is None: si = self.base.si
        if si_bra is None: si_bra = si[:,bra]
        if si_ket is None: si_ket = si[:,ket]
        nroots = self.nroots
        log = lib.logger.new_logger (self, verbose)
        g_model = np.multiply.outer (si_bra.conj (), si_ket)
        g_model -= g_model.T
        g_model *= self.base.e_states[bra]-self.base.e_states[ket]
        g_model = g_model[np.tril_indices (nroots, k=-1)]
        log.debug ("NACs g_is additional component:\n{}".format (g_model))
        return self.pack_uniq_var (g_orb, g_ci, g_is+g_model)

    def get_ham_response (self, **kwargs):
        nac = mspdft_grad.Gradients.get_ham_response (self, **kwargs)
        use_etfs = kwargs.get ('use_etfs', self.use_etfs)
        if not use_etfs:
            verbose = kwargs.get ('verbose', self.verbose)
            log = lib.logger.new_logger (self, verbose)
            nac_model = self.nac_model (**kwargs)
            log.info ('NACs model-state contribution:\n{}'.format (nac_model))
            nac += nac_model
        return nac

    def nac_model (self, mo_coeff=None, ci=None, si=None, si_bra=None,
                   si_ket=None, state=None, mf_grad=None, atmlst=None,
                   **kwargs):
        if state is None: state = self.state
        ket, bra = _unpack_state (state)
        if mo_coeff is None: mo_coeff = self.base.mo_coeff
        if ci is None: ci = self.base.ci
        if si is None: si = self.base.si
        if si_bra is None: si_bra = si[:,bra]
        if si_ket is None: si_ket = si[:,ket]
        if mf_grad is None: mf_grad = self.base.get_rhf_base ().nuc_grad_method ()
        if atmlst is None: atmlst = self.atmlst
        nac = nac_model (self, mo_coeff=mo_coeff, ci=ci, si_bra=si_bra,
                         si_ket=si_ket, mf_grad=mf_grad, atmlst=atmlst)
        e_bra = self.base.e_states[bra]
        e_ket = self.base.e_states[ket]
        nac *= e_bra - e_ket
        return nac

    def kernel (self, *args, **kwargs):
        mult_ediff = kwargs.get ('mult_ediff', self.mult_ediff)
        state = kwargs.get ('state', self.state)
        nac = mspdft_grad.Gradients.kernel (self, *args, **kwargs)
        if not mult_ediff:
            ket, bra = _unpack_state (state)
            e_bra = self.base.e_states[bra]
            e_ket = self.base.e_states[ket]
            nac /= e_bra - e_ket
        return nac

if __name__=='__main__':
    from pyscf import gto, scf
    from mrh.my_pyscf.dft.openmolcas_grids import quasi_ultrafine
    from scipy import linalg
    mol = gto.M (atom = 'Li 0 0 0; H 0 0 1.5', basis='sto-3g',
                 output='mspdft_nacs.log', verbose=lib.logger.INFO)
    mf = scf.RHF (mol).run ()
    mc = mcpdft.CASSCF (mf, 'ftLDA,VWN3', 2, 2, grids_attr=quasi_ultrafine)
    mc = mc.fix_spin_(ss=0).multi_state ([0.5,0.5], 'cms').run (conv_tol=1e-10)
    #openmolcas_energies = np.array ([-7.85629118, -7.72175252])
    print ("energies:",mc.e_states)
    #print ("disagreement w openmolcas:", np.around (mc.e_states-openmolcas_energies, 8))
    mc_nacs = NonAdiabaticCouplings (mc)
    print ("no csf contr")
    print ("antisym")
    nac_01 = mc_nacs.kernel (state=(0,1), use_etfs=True)
    nac_10 = mc_nacs.kernel (state=(1,0), use_etfs=True)
    print (nac_01)
    print (nac_10)
    print ("checking antisym:",linalg.norm(nac_01+nac_10))
    print ("sym")
    nac_01_mult = mc_nacs.kernel (state=(0,1), use_etfs=True, mult_ediff=True)
    nac_10_mult = mc_nacs.kernel (state=(1,0), use_etfs=True, mult_ediff=True)
    print (nac_01_mult)
    print (nac_10_mult)
    print ("checking sym:",linalg.norm(nac_01_mult-nac_10_mult))


    print ("incl csf contr")
    print ("antisym")
    nac_01 = mc_nacs.kernel (state=(0,1), use_etfs=False)
    nac_10 = mc_nacs.kernel (state=(1,0), use_etfs=False)
    print (nac_01)
    print ("checking antisym:",linalg.norm(nac_01+nac_10))
    print ("sym")
    nac_01_mult = mc_nacs.kernel (state=(0,1), use_etfs=False, mult_ediff=True)
    nac_10_mult = mc_nacs.kernel (state=(1,0), use_etfs=False, mult_ediff=True)
    print (nac_01_mult)
    print ("checking sym:",linalg.norm(nac_01_mult-nac_10_mult))

    print ("Check gradients")
    mc_grad = mc.nuc_grad_method ()
    de_0 = mc_grad.kernel (state=0)
    print (de_0)
    de_1 = mc_grad.kernel (state=1)
    print (de_1)

    #from mrh.my_pyscf.tools.molcas2pyscf import *
    #mol = get_mol_from_h5 ('LiH_sa2casscf22_sto3g.rasscf.h5',
    #                       output='sacasscf_nacs_fromh5.log',
    #                       verbose=lib.logger.INFO)
    #mo = get_mo_from_h5 (mol, 'LiH_sa2casscf22_sto3g.rasscf.h5')
    #nac_etfs_ref = np.array ([9.14840490109073E-02, -9.14840490109074E-02])
    #nac_ref = np.array ([1.83701578323929E-01, -6.91459741744125E-02])
    #mf = scf.RHF (mol).run ()
    #mc = mcscf.CASSCF (mol, 2, 2).fix_spin_(ss=0).state_average ([0.5,0.5])
    #mc.run (mo, natorb=True, conv_tol=1e-10)
    #mc_nacs = NonAdiabaticCouplings (mc)
    #nac = mc_nacs.kernel (state=(0,1))
    #print (nac)
    #print (nac_ref)
    #nac_etfs = mc_nacs.kernel (state=(0,1), use_etfs=True)
    #print (nac_etfs)
    #print (nac_etfs_ref)





#!/usr/bin/env python
# Copyright 2019 The PySCF Developers. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
# Author: Qiming Sun <osirpt.sun@gmail.com>
#

'''
Mole class for MM particles
'''

import numpy
from pyscf import gto
from pyscf.gto.mole import is_au
from pyscf.data.elements import charge
from pyscf.lib import param

class Mole(gto.Mole):
    '''Mole class for MM particles.

    Args:
        atoms : geometry of MM particles (unit Bohr).

            | [[atom1, (x, y, z)],
            |  [atom2, (x, y, z)],
            |  ...
            |  [atomN, (x, y, z)]]

    Kwargs:
        charges : 1D array
            fractional charges of MM particles
        zeta : 1D array
            Gaussian charge distribution parameter.
            rho(r) = charge * Norm * exp(-zeta * r^2)

    '''
    def __init__(self, atoms, charges=None, zeta=None):
        gto.Mole.__init__(self)
        self.atom = self._atom = atoms
        self.unit = 'Bohr'
        self.charge_model = 'point'

        # Initialize ._atm and ._env to save the coordinates and charges and
        # other info of MM particles
        natm = len(atoms)
        _atm = numpy.zeros((natm,6), dtype=numpy.int32)
        _atm[:,gto.CHARGE_OF] = [charge(a[0]) for a in atoms]
        coords = numpy.asarray([a[1] for a in atoms], dtype=numpy.double)
        if charges is None:
            _atm[:,gto.NUC_MOD_OF] = gto.NUC_POINT
            charges = _atm[:,gto.CHARGE_OF:gto.CHARGE_OF+1]
        else:
            _atm[:,gto.NUC_MOD_OF] = gto.NUC_FRAC_CHARGE
            charges = numpy.asarray(charges)[:,numpy.newaxis]

        self._env = numpy.append(numpy.zeros(gto.PTR_ENV_START),
                                 numpy.hstack((coords, charges)).ravel())
        _atm[:,gto.PTR_COORD] = gto.PTR_ENV_START + numpy.arange(natm) * 4
        _atm[:,gto.PTR_FRAC_CHARGE] = gto.PTR_ENV_START + numpy.arange(natm) * 4 + 3

        if zeta is not None:
            self.charge_model = 'gaussian'
            zeta = numpy.asarray(zeta, dtype=float).ravel()
            self._env = numpy.append(self._env, zeta)
            _atm[:,gto.PTR_ZETA] = gto.PTR_ENV_START + natm*4 + numpy.arange(natm)

        self._atm = _atm
        self._built = True

    def get_zetas(self):
        if self.charge_model == 'gaussian':
            return self._env[self._atm[:,gto.PTR_ZETA]]
        else:
            return 1e16

def create_mm_mol(atoms_or_coords, charges=None, radii=None, unit='Angstrom'):
    '''Create an MM object based on the given coordinates and charges of MM
    particles.

    Args:
        atoms_or_coords : array-like
            Cartesian coordinates of MM atoms, in the form of a 2D array:
            [(x1, y1, z1), (x2, y2, z2), ...]

    Kwargs:
        charges : 1D array
            The charges of MM atoms.
        radii : 1D array
            The Gaussian charge distribution radii of MM atoms.
        unit : string
            The unit of the input. Default is 'Angstrom'.
    '''
    if isinstance(atoms_or_coords, numpy.ndarray):
        # atoms_or_coords == np.array([(xx, xx, xx)])
        # Patch ghost atoms
        atoms = [(0, c) for c in atoms_or_coords]
    elif (isinstance(atoms_or_coords, (list, tuple)) and
          atoms_or_coords and
          isinstance(atoms_or_coords[0][1], (int, float))):
        # atoms_or_coords == [(xx, xx, xx)]
        # Patch ghost atoms
        atoms = [(0, c) for c in atoms_or_coords]
    else:
        atoms = atoms_or_coords
    atoms = gto.format_atom(atoms, unit=unit)

    if radii is None:
        zeta = None
    else:
        radii = numpy.asarray(radii, dtype=float).ravel()
        if not is_au(unit):
            radii = radii / param.BOHR
        zeta = 1 / radii**2
    return Mole(atoms, charges=charges, zeta=zeta)
